import io
from functools import cached_property
from typing import Any, List, Literal, Optional

import lmdb
import torch

from .base import DATASETS, BaseDataset

# TODO: update


@DATASETS.register_module()
class LmdbDataset(BaseDataset[bytes]):

    def __init__(
        self,
        *args,
        filepath: str,
        db: Optional[str] = None,
        decoder: Optional[Literal['None', 'pytorch']] = 'pytorch',
        **kwargs,
    ):
        self._env: lmdb.Environment = lmdb.open(
            filepath, readonly=True, max_dbs=1
        )
        self._db: Optional[lmdb._Database] = (  # yapf: disable
            None if db is None else self._env.open_db(db.encode())
        )
        self._decoder = decoder
        super().__init__(*args, **kwargs)

    @classmethod
    def load_from(cls, source: BaseDataset, *args, **kwargs):
        pass

    def _map_indices(self) -> List[bytes]:
        with self.begin() as txn, txn.cursor() as cur:
            return list(cur.iternext(keys=True, values=False))

    @cached_property
    def _len(self) -> int:
        with self.begin() as txn:
            return txn.stat()['entries']

    def _getitem(self, index: bytes) -> Any:
        if not isinstance(index, bytes):
            index = str(index).encode()
        with self.begin() as txn:
            buffer = txn.get(index)
        if self._decoder is None or self._decoder == 'None':
            return buffer
        if self._decoder == 'pytorch':
            buffer = io.BytesIO(buffer)
            return torch.load(buffer, map_location='cpu')
        raise Exception

    def begin(self) -> lmdb.Transaction:
        return self._env.begin(self._db)



__all__ = [
    'L1Loss',
    'MSELoss',
]

from abc import abstractmethod
from typing import Optional

import torch
import torch.nn.functional as F

from .base import LOSSES, BaseLoss


class FunctionalLoss(BaseLoss):

    @staticmethod
    @abstractmethod
    def func(*args, **kwargs) -> torch.Tensor:
        pass

    def forward(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        if mask is None:
            loss = self.func(
                pred,
                target,
                *args,
                reduction=self.reduction,
                **kwargs,
            )
        else:
            loss = self.func(pred, target, *args, reduction='none', **kwargs)
            loss = self.reduce(loss, mask)
        return loss


class NormMixin(FunctionalLoss):

    def __init__(self, *args, norm: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self._norm = norm

    def forward(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        if self._norm:
            pred = F.normalize(pred)
            target = F.normalize(target)
        return super().forward(pred, target, *args, **kwargs)


@LOSSES.register_module()
class L1Loss(NormMixin, FunctionalLoss):

    @staticmethod
    def func(*args, **kwargs) -> torch.Tensor:
        return F.l1_loss(*args, **kwargs)


@LOSSES.register_module()
class MSELoss(NormMixin, FunctionalLoss):

    @staticmethod
    def func(*args, **kwargs) -> torch.Tensor:
        return F.mse_loss(*args, **kwargs)


@LOSSES.register_module()
class BCELoss(FunctionalLoss):

    @staticmethod
    def func(*args, **kwargs) -> torch.Tensor:
        return F.binary_cross_entropy(*args, **kwargs)


@LOSSES.register_module()
class BCEWithLogitsLoss(FunctionalLoss):

    @staticmethod
    def func(*args, **kwargs) -> torch.Tensor:
        return F.binary_cross_entropy_with_logits(*args, **kwargs)


@LOSSES.register_module()
class CrossEntropyLoss(FunctionalLoss):

    @staticmethod
    def func(*args, **kwargs) -> torch.Tensor:
        return F.cross_entropy(*args, **kwargs)



__all__ = [
    'L12DLoss',
    'MSE2DLoss',
    'FGDLoss',
    'FGFILoss',
]

from typing import Optional

import einops
import torch
import torch.nn.functional as F

from .base import LOSSES
from .functional import FunctionalLoss, L1Loss, MSELoss


class _2DMixin(FunctionalLoss):

    def forward(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        _, _, h, w = pred.shape
        if pred.shape != target.shape:
            target = F.adaptive_avg_pool2d(target, (h, w))
        if mask is not None and pred.shape != mask.shape:
            mask = F.adaptive_avg_pool2d(mask, (h, w))
        return super().forward(pred, target, mask, *args, **kwargs)


@LOSSES.register_module()
class L12DLoss(_2DMixin, L1Loss):
    pass


@LOSSES.register_module()
class MSE2DLoss(_2DMixin, MSELoss):
    pass


@LOSSES.register_module()
class FGFILoss(MSE2DLoss):

    def forward(  # type: ignore[override]
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        mask: torch.Tensor,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        n, _, h, w = pred.shape
        target = F.adaptive_avg_pool2d(target, (h, w))
        assert mask.shape == (n, 1, h, w) and mask.dtype == torch.bool

        pred = einops.rearrange(pred, 'n c h w -> n h w c')
        target = einops.rearrange(target, 'n c h w -> n h w c')
        mask = einops.rearrange(mask, 'n 1 h w -> n h w')

        return super().forward(  # type: ignore[misc]
            pred[mask],
            target[mask],
            mask=None,
            *args,
            **kwargs,
        )


@LOSSES.register_module()
class FGDLoss(MSE2DLoss):

    def forward(  # type: ignore[override]
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        attn_spatial: torch.Tensor,
        attn_channel: torch.Tensor,
        mask: torch.Tensor,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        _, _, h, w = pred.shape
        attn_spatial = F.adaptive_avg_pool2d(attn_spatial, (h, w))
        attn_channel = F.adaptive_avg_pool2d(attn_channel, (h, w))
        mask = attn_spatial * attn_channel * mask
        return super().forward(  # type: ignore[misc]
            pred,
            target,
            mask=mask,
            *args,
            **kwargs,
        )


# @LOSSES.register_module()
# class LabelEncLoss(MSE2DLoss):

#     def __init__(
#         self,
#         *args,
#         num_channels: int,
#         weight: float = 1.0,
#         **kwargs,
#     ):
#         weight = LinearScheduler(
#             start_value=0,
#             end_value=weight,
#             start_iter=30000,
#             end_iter=30000,
#         )
#         super().__init__(*args, weight=weight, **kwargs)

#         norm_cfg = dict(type='GN', num_groups=1, affine=False)
#         self._adapt = Sequential(
#             ConvModule(num_channels, num_channels, 3, 1, 1),
#             ConvModule(num_channels, num_channels, 3, 1, 1),
#             ConvModule(
#                 num_channels,
#                 num_channels,
#                 3,
#                 1,
#                 1,
#                 norm_cfg=norm_cfg,
#                 act_cfg=None,
#             ),
#         )
#         self._target_norm: nn.Module = NORM_LAYERS.build(
#             norm_cfg,
#             default_args=dict(num_features=num_channels),
#         )

#     def forward(  # type: ignore[override]
#         self,
#         preds: List[torch.Tensor],
#         targets: List[torch.Tensor],
#         *args,
#         **kwargs,
#     ) -> List[torch.Tensor]:
#         losses = []
#         for pred, target in zip(preds, targets):
#             pred = self._adapt(pred)
#             target = self._target_norm(target.detach())
#             loss = super().forward(pred, target, *args, **kwargs)
#             losses.append(loss)
#         return losses

# @LOSSES.register_module()
# class DevLoss(MSE2DLoss):
#     def __init__(self, pos_share: float = 1, *args, **kwargs):
#         assert 0 <= pos_share <= 1
#         super().__init__(*args, **kwargs)
#         self.pos_share = pos_share

#     def pos_mask(self, mask: torch.Tensor) -> torch.Tensor:
#         pos = mask.sum().item()
#         total = mask.numel()
#         pos_frac = self.pos_share * total / pos if pos > 0 else 0
#         neg_frac = (
#             (1 - self.pos_share)  * total / (total - pos)
#             if pos < total else 0
#         )
#         weight = torch.zeros_like(mask, dtype=torch.float)
#         weight.masked_fill_(mask, pos_frac)
#         weight.masked_fill_(~mask, neg_frac)
#         return weight

#     def iou_mask(
#         self, mask: torch.Tensor, pos_thresh: float = 0.5, index: float = 2,
#     ) -> torch.Tensor:
#         """Adjust IoU mask to satisfy `pos_share` requirement.

#         Adjust IoU mask to retrieve `weight` w such that
#         $$
#             w = a * mask^index + b, \\quad a, b \\in \\R^+
#             w_p = w * (w > pos_thresh)
#             sum(w) = mask.numel()
#             sum(w_p) = sum(w) * self.pos_share
#         $$

#         The equations are sometimes unsatisfiable.
#         The algorithms goes briefly as follows:
#         1. estimate b' to balance pos and neg
#         2. normalize sum of weight to `mask.numel()`
#         """
#         assert mask.requires_grad == False
#         n_total = mask.numel()
#         pos_mask = mask > pos_thresh
#         n_pos = pos_mask.sum()
#         mask = mask ** index
#         total = mask.sum()
#         pos = mask[pos_mask].sum()
#         base = (
#             (pos / self.pos_share - total)
#             / (n_total - n_pos / self.pos_share)
#         )
#         base.clamp_(1e-5, 1)
#         mask += base
#         weight = mask * n_total / mask.sum()
#         return weight

#     def forward(
#         self, pred: torch.Tensor, target: torch.Tensor,
#         mask: torch.Tensor, *args, **kwargs,
#     ):
#         mask = mask.unsqueeze(1)  # bs x 1 x h x w
#         if mask.dtype == torch.bool:
#             mask = self.pos_mask(mask)
#         elif mask.dtype == torch.float:
#             mask = self.iou_mask(mask)
#         else:
#             mask = torch.ones_like(pred)

#         if mask.shape != pred.shape:
#             mask = F.interpolate(mask, size=pred.shape[-2:])

#         return super().forward(pred, target, *args, mask=mask, **kwargs)



__all__ = [
    'BaseHook',
    'HOOKS',
]

from abc import abstractmethod
from enum import IntEnum, auto
from typing import TYPE_CHECKING, Protocol

import torch.nn as nn

from ..base import STEPS, Registry, StatusMixin, getattr_recur

if TYPE_CHECKING:

    class _HookProto(Protocol):
        _path: str

        def register_tensor(self, tensor) -> None:
            pass

else:
    _HookProto = object


class Status(IntEnum):
    INITIALIZED = auto()
    BINDED = auto()
    REGISTERED = auto()


class _HookMixin(_HookProto):

    def _forward_hook(self, module: nn.Module, input_, output) -> None:
        self.register_tensor(output)

    def bind(self, model: nn.Module) -> None:
        module: nn.Module = getattr_recur(model, self._path)
        module.register_forward_hook(self._forward_hook)


class _TrackingMixin(_HookProto):

    def __init__(
        self,
        tracking_mode: bool = False,
    ) -> None:
        self._tracking_mode = tracking_mode

    @property
    def tracking_mode(self) -> bool:
        return self._tracking_mode

    def bind(self, model: nn.Module) -> None:
        self._model = model

    def track_tensor(self) -> None:
        if not self._tracking_mode:
            raise AttributeError(f"Hook {self._path} does not support track.")
        tensor = getattr_recur(self._model, self._path)
        self.register_tensor(tensor)


class BaseHook(StatusMixin[Status], _HookMixin, _TrackingMixin):

    def __init__(
        self,
        path: str,
        tracking_mode: bool = False,
    ) -> None:
        StatusMixin.__init__(self, Status.INITIALIZED)
        _HookMixin.__init__(self)
        _TrackingMixin.__init__(self, tracking_mode)
        self._path = path
        self._reset()

    def __call__(self):
        if self.status != Status.REGISTERED:
            raise RuntimeError(f"Hook {self._path} is not registered.")
        return self._tensor()

    @property
    def path(self) -> str:
        return self._path

    @abstractmethod
    def _reset(self):
        pass

    @abstractmethod
    def _register_tensor(self, tensor) -> None:
        pass

    @abstractmethod
    def _tensor(self):
        pass

    @StatusMixin.transit(
        (Status.BINDED, Status.REGISTERED),
        Status.BINDED,
    )
    def reset(self) -> None:
        self._reset()

    @StatusMixin.transit(Status.INITIALIZED, Status.BINDED)
    def bind(self, model: nn.Module) -> None:
        if self._tracking_mode:
            _TrackingMixin.bind(self, model)
        else:
            _HookMixin.bind(self, model)

    @StatusMixin.transit(
        (Status.BINDED, Status.REGISTERED),
        Status.REGISTERED,
    )
    def register_tensor(self, tensor) -> None:
        self._register_tensor(tensor)


HOOKS: Registry[BaseHook] = Registry(
    'hooks',
    parent=STEPS,
    base=BaseHook,
)



__all__ = [
    'ckd_loss',
    'CKDLoss',
]

from typing import List, Optional, Tuple

import torch
import torch.distributed as dist
import torch.nn.functional as F

from ..base import BBoxesXYXY, get_rank, get_world_size
from .base import LOSSES, BaseLoss


def ckd_loss(
    pred: torch.Tensor,
    target: torch.Tensor,
    ignore: Optional[Tuple[torch.Tensor, torch.Tensor]] = None,
    gamma: float = 0.07,
) -> torch.Tensor:
    """Warpper of CKD loss.

    Refer to http://arxiv.org/abs/2108.07482.

    Args:
        pred: n x dim
            Normalized predictions.

        target: m x dim
            Normalized targets. First `n` targets correspond to `pred`.

        ignore: (n, n)

    Returns:
        loss: 1
    """
    if ignore is None:
        inds = torch.arange(
            pred.shape[0],
            dtype=torch.long,
            device=pred.device,
        )
        ignore = (inds, ) * 2
    similarity = target.mm(pred.t()) / gamma  # m x n
    similarity = similarity.exp()
    pos = torch.diag(similarity)  # n
    similarity = similarity.index_put(ignore, similarity.new_zeros(()))
    total = similarity.sum(0)  # n
    loss = pos / total  # n
    return -loss.log().mean()


class MemoryPool:

    def __init__(self, size: int = 10):
        self._memory: List[torch.Tensor] = []
        self._size = size
        self._rank = get_rank()
        self._world_size = get_world_size()

    def register(self, tensor: torch.Tensor) -> None:
        tensor = tensor.contiguous()
        if self._world_size > 1:
            tensor_list = [
                torch.zeros_like(tensor) for _ in range(self._world_size)
            ]
            dist.all_gather(tensor_list, tensor)
            tensor_list[0], tensor_list[self._rank] = \
                tensor_list[self._rank], tensor_list[0]
            tensor = torch.cat(tensor_list)
        else:
            tensor = tensor.detach()
        self._memory.insert(0, tensor)
        if len(self._memory) > self._size:
            self._memory.pop(-1)

    @property
    def memory(self) -> torch.Tensor:
        return torch.cat(self._memory)


@LOSSES.register_module()
class CKDLoss(BaseLoss):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._memory_pool = MemoryPool()

    def forward(
        self,
        preds: torch.Tensor,
        targets: torch.Tensor,
        bboxes: Optional[List[torch.Tensor]] = None,
    ) -> torch.Tensor:
        """Compute CKD loss.

        Refer to http://arxiv.org/abs/2108.07482.

        Args:
            preds: m x dim
            targets: m x dim
            bboxes: n x m x 4

        Returns:
            loss: 1
        """
        assert preds.shape == targets.shape, (preds.shape, targets.shape)
        preds = F.normalize(preds)
        targets = F.normalize(targets)
        self._memory_pool.register(targets)

        if bboxes is None:
            loss = ckd_loss(preds, self._memory_pool.memory)
        else:
            ignore_x, ignore_y, ind = [], [], 0
            for bbox in bboxes:
                bbox_ = BBoxesXYXY(bbox)
                ious = BBoxesXYXY.ious(bbox_, bbox_)
                x, y = torch.where(ious > 0.5)
                ignore_x.append(x + ind)
                ignore_y.append(y + ind)
                ind += len(bbox_)
            assert ind == preds.shape[0], (ind, preds.shape)
            ignore = (torch.cat(ignore_x), torch.cat(ignore_y))
            loss = ckd_loss(preds, self._memory_pool.memory, ignore)

        return loss



import os
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Any, Iterator

import torch

from .base import ACCESS_LAYERS, DATASETS, BaseAccessLayer, BaseDataset, Codec


@ACCESS_LAYERS.register_module()
class ZipAccessLayer(BaseAccessLayer[str]):

    def _init(self):
        assert self._readonly, 'ZipAccessLayer is read-only.'
        assert self._codec is Codec.PYTORCH, \
            'ZipAccessLayer only supports PyTorch.'
        data_root_parts = Path(self._data_root).parts
        zip_indices = [  # yapf: disable
            i for i, part in enumerate(data_root_parts)
            if part.endswith('.zip')
        ]
        if len(zip_indices) == 0:
            raise ValueError(f'{self._data_root} does not contain a zip file.')
        elif len(zip_indices) > 1:
            raise ValueError(f'{self._data_root} contains multiple zip files.')
        zip_index = zip_indices[0]
        self._zip_root = zipfile.Path(
            os.path.join(*data_root_parts[:zip_index + 1]),
            os.path.join(
                *data_root_parts[zip_index + 1:],
                self._task_name,
                '',
            ),
        )

    @property
    def exists(self) -> bool:
        return self._zip_root.exists()

    def touch(self):
        raise NotImplementedError

    def __iter__(self) -> Iterator[str]:
        return (path.name for path in self._zip_root.iterdir())

    def __len__(self) -> int:
        return len(list(self._zip_root.iterdir()))

    def __getitem__(self, key: str) -> Any:
        data_file = self._zip_root / key
        bytes_io = BytesIO(data_file.read_bytes())
        return torch.load(bytes_io, map_location='cpu')

    def __setitem__(self, key: str, value: Any):
        raise NotImplementedError

    def __delitem__(self, key: str):
        raise NotImplementedError


@DATASETS.register_module()
class ZipDataset(BaseDataset[str]):
    ACCESS_LAYER = ZipAccessLayer



import sys
import typing

from typing_extensions import Literal, Protocol, runtime_checkable

from .._extensions import get_logger

if sys.version_info < (3, 8):
    get_logger().warning(
        "Monkey patching `typing.Literal`, `typing.Protocol`, "
        "and `typing.runtime_checkable`.",
    )
    typing.Literal = Literal
    typing.Protocol = Protocol
    typing.runtime_checkable = runtime_checkable



__all__ = [
    'Module',
    'ModuleList',
    'ModuleDict',
    'Sequential',
    'get_rank',
    'get_world_size',
]

import torch.distributed as dist

try:
    from mmcv.runner import BaseModule as Module
    from mmcv.runner import ModuleDict, ModuleList, Sequential
except Exception:
    from torch.nn import Module, ModuleDict, ModuleList, Sequential


def get_rank(*args, **kwargs) -> int:
    if dist.is_available() and dist.is_initialized():
        return dist.get_rank(*args, **kwargs)
    return 0


def get_world_size(*args, **kwargs) -> int:
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size(*args, **kwargs)
    return 1



import math
from abc import abstractmethod
from typing import List, Tuple

import einops
import torch

from .base import ADAPTS, BaseAdapt


class MultiLevelMask:

    def __init__(
        self,
        *args,
        strides: List[int],
        ceil_mode: bool = False,
        **kwargs,
    ):
        self._strides = strides
        self._ceil_mode = ceil_mode
        super().__init__(*args, **kwargs)

    @property
    def strides(self) -> List[int]:
        return self._strides

    @abstractmethod
    def _forward(
        self,
        shape: Tuple[int, int],
        bboxes: List[torch.Tensor],
        *args,
        **kwargs,
    ) -> torch.Tensor:
        pass

    def _div(self, a: int, b: int) -> int:
        if self._ceil_mode:
            return math.ceil(a / b)
        return a // b

    def forward(
        self,
        shape: Tuple[int, int],
        bboxes: List[torch.Tensor],
        *args,
        **kwargs,
    ) -> List[torch.Tensor]:
        """
        Args:
            shape: (h, w)
            bboxes: n x m x 4

        Returns:
            masks: l x n x 1 x h x w
        """
        h, w = shape
        masks = []
        for stride in self._strides:
            level_shape = (self._div(h, stride), self._div(w, stride))
            level_bboxes = [bbox / stride for bbox in bboxes]
            mask = self._forward(level_shape, level_bboxes, *args, **kwargs)
            masks.append(mask)
        return masks


class SingleLevelMask(MultiLevelMask):

    def __init__(self, *args, stride: int, **kwargs):
        super().__init__(*args, strides=[stride], **kwargs)

    @property
    def stride(self) -> int:
        return self.strides[0]

    def forward(  # type: ignore[override]
        self,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        masks = super().forward(*args, **kwargs)
        return masks[0]


@ADAPTS.register_module()
class LabelEncMask(BaseAdapt):

    def __init__(
        self,
        *args,
        num_classes: int = 80,
        aug: bool = True,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._num_classes = num_classes
        self._aug = aug

    def _mask(
        self,
        shape: Tuple[int, int],
        bboxes: torch.Tensor,
        labels: torch.Tensor,
    ) -> torch.Tensor:
        """
        Args:
            shape: (h, w)
            bboxes: m x 4
            labels: m

        Returns:
            mask: k x h x w
        """
        masks = bboxes.new_zeros(self._num_classes, *shape)
        bboxes = torch.cat(
            [
                bboxes[:, 2:] - bboxes[:, :2],
                (bboxes[:, :2] + bboxes[:, 2:]) / 2
            ],
            dim=-1,
        )
        y, x = torch.meshgrid(
            torch.arange(0, shape[0], dtype=torch.float, device=bboxes.device),
            torch.arange(0, shape[1], dtype=torch.float, device=bboxes.device),
        )
        for (w, h, cx, cy), label in zip(bboxes.tolist(), labels.tolist()):
            value = torch.max(
                torch.abs(x - cx) / w,
                torch.abs(y - cy) / h,
            )
            value = (1 - value) * (value < 0.5)
            if self._aug:
                weight = torch.rand((), device=value.device).clamp_max(0.5) * 2
                value = value * weight
            torch.maximum(masks[label], value, out=masks[label])
        return masks

    def forward(
        self,
        shape: Tuple[int, int],
        bboxes: List[torch.Tensor],
        labels: List[torch.Tensor],
    ) -> torch.Tensor:
        """
        Args:
            shape: (h, w)
            bboxes: n x m x 4
            labels: n x m

        Returns:
            masks: n x k x h x w
        """
        masks = [  # yapf: disable
            self._mask(shape, bbox, label)
            for bbox, label in zip(bboxes, labels)
        ]
        return torch.stack(masks)


@ADAPTS.register_module()
class DeFeatMask(MultiLevelMask, BaseAdapt):

    def __init__(
        self,
        *args,
        neg_gain: float = 4,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._neg_gain = neg_gain

    @staticmethod
    def _normalize(masks: torch.Tensor) -> torch.Tensor:
        """
        Args:
            masks: n x 1 x h x w

        Returns:
            normalized_masks: n x 1 x h x w
        """
        values = einops.reduce(
            masks,
            'n 1 h w -> n 1 1 1',
            reduction='sum',
        ).clamp_min_(1)
        normalized_masks = torch.true_divide(masks, values)
        return normalized_masks

    def _normalize_pos(self, masks: torch.Tensor) -> torch.Tensor:
        return self._normalize(masks)

    def _normalize_neg(self, masks: torch.Tensor) -> torch.Tensor:
        return self._normalize(masks)

    def _mask(
        self,
        shape: Tuple[int, int],
        bboxes: torch.Tensor,
    ) -> torch.Tensor:
        """
        Args:
            shape: (h, w)
            bboxes: m x 4

        Returns:
            mask: h x w
        """
        mask = bboxes.new_zeros(shape)
        for x0, y0, x1, y1 in bboxes.int().tolist():
            mask[y0:y1 + 2, x0:x1 + 2] = 1
        return mask

    def _forward(
        self,
        shape: Tuple[int, int],
        bboxes: List[torch.Tensor],
        *args,
        **kwargs,
    ) -> torch.Tensor:
        """
        Args:
            shape: (h, w)
            bboxes: n x m x 4

        Returns:
            masks: n x 1 x h x w
        """
        masks = torch.stack([self._mask(shape, bbox) for bbox in bboxes])
        masks = einops.rearrange(masks, 'n h w -> n 1 h w')
        masks = self._normalize_pos(masks)
        neg_masks = self._normalize_neg(masks <= 0)
        neg_masks = neg_masks * self._neg_gain
        return masks + neg_masks


@ADAPTS.register_module()
class FGDMask(DeFeatMask):

    def _normalize_pos(self, masks: torch.Tensor) -> torch.Tensor:
        return masks

    def _mask(
        self,
        shape: Tuple[int, int],
        bboxes: torch.Tensor,
    ) -> torch.Tensor:
        """
        Args:
            shape: (h, w)
            bboxes: m x 4

        Returns:
            mask: h x w
        """
        mask = bboxes.new_zeros(shape)
        bboxes = bboxes.int()
        values = torch.true_divide(
            1.0,
            (bboxes[:, 2:] - bboxes[:, :2] + 2).prod(1),
        )
        for i, (x0, y0, x1, y1) in enumerate(bboxes.tolist()):
            area = mask[y0:y1 + 2, x0:x1 + 2]
            torch.maximum(area, values[i], out=area)
        return mask


@ADAPTS.register_module()
class FGFIMask(BaseAdapt):

    def __init__(self, *args, thresh: float = 0.5, **kwargs):
        super().__init__(*args, **kwargs)
        self._thresh = thresh

    def _instance(self, ious: torch.Tensor) -> torch.Tensor:
        """
        Args:
            ious: h x w x k x m

        Returns:
            mask: h x w
        """
        thresh = einops.reduce(
            ious,
            'h w k m -> 1 1 1 m',
            reduction='max',
        ) * self._thresh
        mask = einops.reduce(ious > thresh, 'h w k m -> h w', reduction='max')
        return mask

    def _batch(self, ious: List[torch.Tensor]) -> torch.Tensor:
        """
        Args:
            ious: n x h x w x k x m

        Returns:
            masks: n x 1 x h x w
        """
        masks = torch.stack([self._instance(iou) for iou in ious])
        masks = einops.rearrange(masks, 'n h w -> n 1 h w')
        return masks

    def forward(self, ious: List[List[torch.Tensor]]) -> List[torch.Tensor]:
        """
        Args:
            ious: l x n x h x w x k x m

        Returns:
            masks: l x n x 1 x h x w
        """
        masks = [self._batch(iou) for iou in ious]
        return masks


# @ADAPTS.register_module()
# class DenseCLIPMask(SingleLevelMask, LabelEncMask):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, stride=32, aug=False, **kwargs)

#     def _mask(
#         self,
#         shape: Tuple[int, int],
#         bboxes: torch.Tensor,
#         labels: torch.Tensor,
#     ) -> torch.Tensor:
#         masks = bboxes.new_zeros(self._num_classes, *shape)
#         for (x0, y0, x1, y1), label in zip(
#             bboxes.int().tolist(),
#             labels.tolist(),
#         ):
#             masks[label, y0:y1 + 1, x0:x1 + 1] = 1
#         return masks


@ADAPTS.register_module()
class FRSMask(BaseAdapt):

    def __init__(self, *args, with_logits: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self._with_logits = with_logits

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: n x c x h x w

        Returns:
            masks: n x 1 x h x w
        """
        x = x.detach()
        masks = einops.reduce(x, 'n c h w -> n 1 h w', reduction='max')
        if self._with_logits:
            masks = masks.sigmoid()
        return masks



from typing import Dict, Optional

from ..base import Module, Registry, getattr_recur


class ModelLoader:

    @staticmethod
    def load_mmlab_models(
        registry: Registry[Module],
        config: str,
        config_options: Optional[str] = None,
        ckpt: Optional[str] = None,
    ) -> Module:
        import mmcv
        from mmcv.runner import load_checkpoint

        config_dict = mmcv.Config.fromfile(config)
        if config_options is not None:
            config_dict.merge_from_dict(config_options)
        model = registry.build(config_dict.model)
        if ckpt is not None:
            load_checkpoint(model, ckpt, map_location='cpu')
            model._is_init = True
        return model

    @staticmethod
    def load_state_dict(target: Module, source: Module):
        state_dict = source.state_dict()
        missing_keys, unexpected_keys = target.load_state_dict(
            state_dict, strict=False,
        )
        if len(missing_keys) > 0 or len(unexpected_keys) > 0:
            print(missing_keys, unexpected_keys)  # TODO: enrich output
        target._is_init = True

    @staticmethod
    def load_state_dicts(models, prefixes: Dict[str, str]):
        for target, source in prefixes.items():
            target_module = getattr_recur(models, target)
            source_module = getattr_recur(models, source)
            ModelLoader.load_state_dict(target_module, source_module)



__all__ = [
    'SelfDistiller',
]

from typing import Dict, Optional

import torch.nn as nn

from ..base import WorkflowConfig
from .base import DISTILLERS, BaseDistiller, DecoratorMixin


@DISTILLERS.register_module()
class SelfDistiller(DecoratorMixin, BaseDistiller):

    def __init__(
        self,
        student: nn.Module,
        student_hooks: WorkflowConfig = None,
        weight_transfer: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        assert 'hooks' not in kwargs
        assert 'trackings' not in kwargs

        hooks = {}
        if student_hooks is not None:
            hooks[0] = student_hooks

        if weight_transfer is not None:
            weight_transfer = {  # yapf: disable
                'models.0.' + k: 'models.0.' + v
                for k, v in weight_transfer.items()
                if k != '' and v != ''
            }

        super().__init__(
            [student],
            hooks=hooks,
            weight_transfer=weight_transfer,
            **kwargs,
        )



__all__ = [
    'PreprocessedDistiller',
]

from typing import Any, Union

from ..base import Module
from ..datasets import ACCESS_LAYERS, BaseAccessLayer
from .base import DISTILLERS
from .teacher import SingleTeacherDistiller


class PreprocessedTeacher(Module):

    def __init__(self, *args, access_layer: BaseAccessLayer, **kwargs):
        super().__init__(*args, **kwargs)
        self._access_layer = access_layer

    def forward(self, key: Any) -> Any:
        return self._access_layer[key]


@DISTILLERS.register_module()
class PreprocessedDistiller(SingleTeacherDistiller):
    teacher: PreprocessedTeacher

    def __init__(
        self,
        *args,
        access_layer: Union[BaseAccessLayer, dict],
        **kwargs,
    ):
        access_layer = ACCESS_LAYERS.build(access_layer)
        teacher = PreprocessedTeacher(access_layer=access_layer)
        super().__init__(  # type: ignore[misc]
            *args,
            teacher=teacher,
            teacher_hooks=None,
            teacher_trackings=None,
            teacher_online=False,
            **kwargs,
        )



__all__ = [
    'strict_zip_len',
]

from typing import Iterable, Sized


def strict_zip_len(iterable: Iterable[Sized]) -> int:
    """Length check for zip.

    Before Python 3.10, zip() stops when the shortest element is exhausted.
    This function checks the length of all elements and raises an error if
    they are not the same.

    Args:
        iterable: Iterable of sized elements.

    Returns:
        Length of the sized elements.

    Raises:
        ValueError: If the length of the sized elements are not the same.
    """
    lens = {len(e) for e in iterable}
    if len(lens) > 1:
        raise ValueError(f'Lengths of iterables are not the same: {lens}')
    return lens.pop()



import hashlib
import random
from contextlib import ContextDecorator
from typing import Any, cast

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist

from ..base import (
    get_iter,
    get_logger,
    get_rank,
    get_world_size,
    iter_initialized,
)


def _randint(high: int = 2**30) -> int:
    seed = np.random.randint(high)
    if get_world_size() > 1:
        tensor = torch.IntTensor(seed if get_rank() == 0 else 0)
        dist.broadcast(tensor, src=0)
        seed = cast(int, tensor.item())
    return seed


def init_seed(
    seed=None,
    deterministic: bool = ...,  # type: ignore[assignment]
):
    if seed is None:
        seed = _randint()
    elif isinstance(seed, int):
        seed %= 2**30
    else:
        if not isinstance(seed, bytes):
            seed = str(seed).encode()
        seed = hashlib.blake2b(seed, digest_size=4).hexdigest()
        seed = int(seed, 16)

    if iter_initialized():
        seed += get_iter()

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    if deterministic is not ...:
        cudnn.deterministic = deterministic
        cudnn.benchmark = not deterministic


class set_seed_temp(ContextDecorator):

    def __init__(
        self,
        seed=None,
        deterministic: bool = ...,  # type: ignore[assignment]
    ):
        if torch.cuda.device_count() > 1:
            get_logger().warn(
                "Seeding is not recommended for multi-GPU training.",
            )
        self._seed = seed
        self._deterministic = deterministic

    def __enter__(self):
        self._random_state = random.getstate()
        self._np_state = np.random.get_state()
        self._torch_state = torch.get_rng_state()
        self._cuda_state = torch.cuda.get_rng_state_all()
        self._prev_detereministic = cudnn.deterministic
        self._prev_benchmark = cudnn.benchmark

        init_seed(self._seed, self._deterministic)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        random.setstate(self._random_state)
        np.random.set_state(self._np_state)
        torch.set_rng_state(self._torch_state)
        torch.cuda.set_rng_state_all(self._cuda_state)
        cudnn.deterministic = self._prev_detereministic
        cudnn.benchmark = self._prev_benchmark



__all__ = [
    'BaseAdapt',
    'ADAPTS',
]

import einops.layers.torch as einops
import torch.nn as nn

from ..base import STEPS, Module, Registry


class BaseAdapt(Module):
    pass


ADAPTS: Registry[nn.Module] = Registry('adapts', parent=STEPS, base=nn.Module)
ADAPTS.register(nn.Linear)
ADAPTS.register(einops.Rearrange)
ADAPTS.register(einops.Reduce)

try:
    from mmcv.cnn import CONV_LAYERS, PLUGIN_LAYERS

    for k, v in PLUGIN_LAYERS.module_dict.items():
        ADAPTS.register(v, name=k)
    for k, v in CONV_LAYERS.module_dict.items():
        ADAPTS.register(v, name=k)
except Exception:
    pass



__all__ = [
    'ChainedScheduler',
]

import math
from typing import Iterable

from .base import SCHEDULERS, BaseScheduler, SchedulerConfig


@SCHEDULERS.register_module()
class ChainedScheduler(BaseScheduler):

    def __init__(self, schedulers: Iterable[SchedulerConfig]) -> None:
        self._schedulers = tuple(map(SCHEDULERS.build, schedulers))

    @property
    def value(self) -> float:
        return math.prod(self._schedulers)



from .base import *
from .pth import PthAccessLayer, PthDataset
from .zip import ZipAccessLayer, ZipDataset

__all__ = [
    'PthAccessLayer',
    'PthDataset',
    'ZipAccessLayer',
    'ZipDataset',
]



__all__ = [
    'DuplicatedHook',
]

from typing import Any, List

from .base import HOOKS
from .standard import StandardHook


@HOOKS.register_module()
class DuplicatedHook(StandardHook):

    def __init__(self, *args, num: int = 1, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._num = num

    def _tensor(self) -> List[Any]:
        return [super()._tensor()] * self._num



__all__ = [
    'BaseLoss',
    'LOSSES',
]

import numbers
from typing import Any, Literal, Optional, Union, cast

import torch
import torch.nn as nn

from ..base import STEPS, Config, Module, Registry
from ..schedulers import SCHEDULERS

Reduction = Literal['none', 'mean', 'sum', 'prod']


class BaseLoss(Module):

    def __init__(
        self,
        reduction: Reduction = 'mean',
        weight: Union[numbers.Real, Config] = 1.0,  # type: ignore[assignment]
        bound: Optional[numbers.Real] = None,
        **kwargs,
    ):
        if not isinstance(weight, numbers.Real):
            weight = cast(numbers.Real, SCHEDULERS.build(weight))
        if bound is not None and bound <= 1e-4:
            raise ValueError('bound must be greater than 1e-4')
        super().__init__(**kwargs)
        self._reduction = reduction

        self._weight = weight
        self._threshold = None if bound is None else bound / weight

        self.register_forward_hook(self._forward_hook)

    @property
    def reduction(self) -> Reduction:
        return self._reduction

    def reduce(
        self,
        loss: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if mask is not None:
            loss = loss * mask
        if self._reduction == 'none':
            pass
        elif self._reduction in ['sum', 'mean', 'prod']:
            loss = getattr(loss, self._reduction)()
        else:
            raise NotImplementedError(self._reduction)
        return loss

    def _forward_hook(
        self,
        module: nn.Module,
        input_: Any,
        output: torch.Tensor,
    ) -> torch.Tensor:
        if self._threshold is None:
            return self._weight * output

        weight = self._weight
        # coef = bound / (weight * output)
        coef = self._threshold / output.item()
        # if bound < weight * output
        if coef < 1.0:
            # weight = bound / output
            weight *= coef

        output = weight * output
        return output


LOSSES: Registry[nn.Module] = Registry(
    'losses',
    parent=STEPS,
    base=nn.Module,
)



__all__ = [
    'DictAction',
]

import argparse
from typing import Any, Dict, Optional, Sequence, Union


class DictAction(argparse.Action):
    """``argparse`` action to parse arguments in the form of key-value pairs.

    Examples:
        >>> parser = argparse.ArgumentParser()
        >>> parser.add_argument('--dict', action=DictAction)
        DictAction(...)
        >>> parser.parse_args('--dict key1:"value1" key2:"value2"'.split())
        Namespace(dict={'key1': 'value1', 'key2': 'value2'})
    """

    def __init__(self, *args, nargs=None, **kwargs):
        """
        Args:
            nargs: The number of dictionary arguments that should be consumed.
        """
        if nargs not in [
            None,
            argparse.OPTIONAL,
            argparse.ZERO_OR_MORE,
            argparse.ONE_OR_MORE,
        ]:
            raise ValueError(f"Invalid nargs={nargs}")

        append = nargs in [argparse.ZERO_OR_MORE, argparse.ONE_OR_MORE]
        required = nargs in [None, argparse.ONE_OR_MORE]
        default = [] if append else None
        super().__init__(
            *args,
            nargs=argparse.ZERO_OR_MORE,
            required=required,
            default=default,
            **kwargs,
        )
        self._append = append

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, Sequence):
            raise ValueError(f'values must be a sequence, but got {values}')
        if not all(isinstance(value, str) for value in values):
            raise ValueError(f'values must be strings, but got {values}')
        value_dict: Dict[str, Any] = {}
        for value in values:
            k, v = value.split(':', 1)
            value_dict[k.strip()] = eval(v)
        if self._append:
            value_dict_list = getattr(namespace, self.dest, [])
            value_dict_list.append(value_dict)
            setattr(namespace, self.dest, value_dict_list)
        else:
            setattr(namespace, self.dest, value_dict)



import itertools
import operator
from functools import reduce
from typing import (
    Iterable,
    Iterator,
    List,
    Mapping,
    Sequence,
    Tuple,
    Union,
    overload,
)

import einops
import torch

from .base import ADAPTS, BaseAdapt

__all__ = [
    'DictTensor',
    'Union_',
    'Intersect',
]

KeyType = Tuple[int, ...]


class DictTensor(Mapping):

    def __init__(
        self,
        keys: Union[KeyType, List[KeyType]],
        values: torch.Tensor,
    ) -> None:
        if not isinstance(keys, list):
            keys = [keys]
        self._keys = {key: i for i, key in enumerate(keys)}
        self._values = values

    @overload
    def __getitem__(self, keys: KeyType) -> torch.Tensor:
        ...

    @overload
    def __getitem__(self, keys: List[KeyType]) -> torch.Tensor:
        ...

    def __getitem__(self, keys):
        inds = (
            self._keys[keys] if not isinstance(keys, list) else
            [self._keys[key] for key in keys]
        )
        return self._values[inds]

    def __len__(self) -> int:
        return len(self._keys)

    def __iter__(self) -> Iterator[KeyType]:
        return iter(self._keys)

    def __contains__(self, keys: object) -> bool:
        if not isinstance(keys, list):
            return keys in self._keys
        return all(key in self._keys for key in keys)

    @classmethod
    def from_tensor(
        cls,
        keys: torch.Tensor,
        values: torch.Tensor,
    ) -> 'DictTensor':
        return DictTensor(list(map(tuple, keys.int().tolist())), values)

    @staticmethod
    def union(
        dict_tensors: Sequence['DictTensor'],
    ) -> Tuple[List['DictTensor'], 'DictTensor']:
        keys = list(set(itertools.chain(*dict_tensors)))
        n = len(keys)
        s = len(dict_tensors)

        values = dict_tensors[0]._values.new_zeros((n, s))
        mask = DictTensor(keys, values)

        union_dict_tensors = []
        for i, dict_tensor in enumerate(dict_tensors):
            inds = [mask._keys[key] for key in dict_tensor._keys]
            mask._values[inds, i] = 1

            shape = (n, ) + dict_tensor._values.shape[1:]
            values = dict_tensor._values.new_zeros(shape)
            values[inds] = dict_tensor._values
            union_dict_tensor = DictTensor(keys, values)

            union_dict_tensors.append(union_dict_tensor)
        return union_dict_tensors, mask

    @staticmethod
    def intersect(dict_tensors: Iterable['DictTensor']) -> List['DictTensor']:
        keys: List[KeyType] = list(
            reduce(operator.and_, map(set, dict_tensors)),
        )
        dict_tensors = [
            DictTensor(keys, dict_tensor[keys]) for dict_tensor in dict_tensors
        ]
        return dict_tensors


@ADAPTS.register_module(name='Union')
class Union_(BaseAdapt):

    def forward(
        self,
        feats: List[torch.Tensor],
        ids: List[torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Match `feats` according to their `poses`.

        Align the `feats` coming from different sources to have same
        `matched_pos` and stack them togethor. For positions where some
        of `feats` do not show up, an all-zero tensor is added as
        default. A 2D `mask` is returned to indicate the type of a
        matched feature, where `1` corresponds to features coming from
        `feats` and `0` for added default all-zero tensors.

        Args:
            feats: [n_s x d_1 x d_2 x ... x d_m]
                Features from `s` different sources, each source can
                have different `n_s`.
            ids: [n_s x l]
                Positions of each feature.

        Returns:
            union_feats: s x n x d_1 x d_2 x ... x d_m
            union_ids: n x l
            union_mask: s x n
        """
        dict_tensors = [
            DictTensor.from_tensor(id_, feat) for id_, feat in zip(ids, feats)
        ]
        dict_tensors, mask = DictTensor.union(dict_tensors)
        union_feats = torch.stack([
            dict_tensor._values for dict_tensor in dict_tensors
        ])
        union_ids = union_feats.new_tensor(mask._keys)
        union_mask = einops.rearrange(mask._values, 'n s -> s n')
        return union_feats, union_ids, union_mask


@ADAPTS.register_module()
class Intersect(BaseAdapt):

    def forward(
        self,
        feats: List[torch.Tensor],
        ids: List[torch.Tensor],
    ) -> List[torch.Tensor]:
        """Match positions that show up both in `pred_poses` and
        `target_poses`.

        Args:
            feats: [n_s x d_1 x d_2 x ... x d_m]
                Features from `s` different sources, each source can
                have different `n_s`.
            ids: [n_s x l]
                Positions of each feature.

        Returns:
            intersect_feats: [n x d_1 x d_2 x ... x d_m]
        """
        dict_tensors = [
            DictTensor.from_tensor(id_, feat) for id_, feat in zip(ids, feats)
        ]
        dict_tensors = DictTensor.intersect(dict_tensors)
        intersect_feats = [dict_tensor._values for dict_tensor in dict_tensors]
        return intersect_feats



__all__ = [
    'iter_initialized',
    'init_iter',
    'get_iter',
    'inc_iter',
]

from ._extensions import get_logger

_iter = None


def iter_initialized() -> bool:
    global _iter
    return _iter is not None


def init_iter(iter_: int = 0):
    global _iter
    if _iter is not None:
        get_logger().warning(f"iter={_iter} has been reset to {iter_}.")
    _iter = iter_


def get_iter() -> int:
    global _iter
    assert _iter is not None
    return _iter


def inc_iter():
    global _iter
    assert _iter is not None
    _iter += 1



__all__ = [
    'StatusMixin',
]

import enum
import functools
from typing import (
    Any,
    Callable,
    Generic,
    Sequence,
    TypeVar,
    Union,
    no_type_check,
)

T = TypeVar('T', bound=enum.Enum)


class StatusMixin(Generic[T]):

    def __init__(self, status: T) -> None:
        self._status = status

    @property
    def status(self) -> T:
        return self._status

    @staticmethod
    def transit(
        source: Union[T, Sequence[T]],
        target: Union[T, Sequence[T]],
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        if not isinstance(source, Sequence):
            assert not isinstance(target, Sequence)
            sources = [source]
            targets = [target]
        elif not isinstance(target, Sequence):
            sources = list(source)
            targets = [target] * len(source)
        else:
            sources = list(source)
            targets = list(target)
        assert len(sources) == len(targets) == len(set(sources)), \
            (source, target)

        @no_type_check
        def wrapper(wrapped_func):

            @functools.wraps(wrapped_func)
            def wrapper_func(self: StatusMixin, *args, **kwargs):
                if self._status not in sources:
                    raise RuntimeError(f"{self} is not in {sources}.")
                self._status = targets[sources.index(self._status)]
                return wrapped_func(self, *args, **kwargs)

            return wrapper_func

        return wrapper



import enum

import torchvision
import torchvision.transforms as transforms
from packaging import version
from PIL import Image

from .._extensions import get_logger

if version.parse(torchvision.__version__) < version.parse('0.9.0'):
    get_logger().warning(
        "Monkey patching `torchvision.transforms.InterpolationMode`.",
    )

    class InterpolationMode(enum.Enum):
        BICUBIC = Image.BICUBIC

    transforms.InterpolationMode = InterpolationMode



import math
from typing import Optional

from .base import SCHEDULERS, BaseScheduler

__all__ = [
    'CosineAnnealingScheduler',
]


@SCHEDULERS.register_module()
class CosineAnnealingScheduler(BaseScheduler):

    def __init__(
        self,
        *args,
        T: int,
        T_mult: int = 1,
        T_max: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._last_iter = 0
        self._T = T
        self._T_mult = T_mult
        self._T_max = T_max

    def _weight(self, cur_iter: int, _) -> float:
        T_cur = cur_iter - self._last_iter
        while T_cur >= self._T:
            T_cur -= self._T
            self._last_iter += self._T
            self._T *= self._T_mult
            if self._T_max is not None and self._T > self._T_max:
                self._T = self._T_max
        return -(1 + math.cos(math.pi * T_cur / self._T)) / 2



from pathlib import Path
from typing import Any, Iterator

import torch

from .base import ACCESS_LAYERS, DATASETS, BaseAccessLayer, BaseDataset, Codec


@ACCESS_LAYERS.register_module()
class PthAccessLayer(BaseAccessLayer[str]):

    def _init(self):
        assert self._codec is Codec.PYTORCH, \
            'PthAccessLayer only supports PyTorch.'
        self._pth_root = Path(self._data_root) / self._task_name

    @property
    def exists(self) -> bool:
        return self._pth_root.exists()

    def touch(self):
        self._pth_root.mkdir(parents=True, exist_ok=self._exist_ok)

    def __iter__(self) -> Iterator[str]:
        return (path.stem for path in self._pth_root.glob('*.pth'))

    def __len__(self) -> int:
        return len(list(self._pth_root.glob('*.pth')))

    def __getitem__(self, key: str) -> Any:
        data_file = self._pth_root / f'{key}.pth'
        return torch.load(data_file, map_location='cpu')

    def __setitem__(self, key: str, value: Any):
        torch.save(value, self._pth_root / f'{key}.pth')

    def __delitem__(self, key: str):
        data_file = self._pth_root / f'{key}.pth'
        data_file.unlink()


@DATASETS.register_module()
class PthDataset(BaseDataset[str]):
    ACCESS_LAYER = PthAccessLayer



import builtins

from .._extensions import get_logger

try:
    import ipdb

    get_logger().info("`ipdb` is installed. Using it for debugging.")
    builtins.breakpoint = ipdb.set_trace
except ImportError:
    pass



from .base import SCHEDULERS, IntervalScheduler

__all__ = [
    'LinearScheduler',
    'ConstantScheduler',
    'WarmupScheduler',
    'EarlyStopScheduler',
    'DecayScheduler',
]


@SCHEDULERS.register_module()
class LinearScheduler(IntervalScheduler):

    def _weight(self, cur_iter: int, total_iter: float) -> float:
        return cur_iter / total_iter


@SCHEDULERS.register_module()
class ConstantScheduler(LinearScheduler):

    def __init__(self, *, value: float = 1) -> None:
        super().__init__(
            start_value=value,
            end_value=value,
            start_iter=-1,
            end_iter=-1,
        )


@SCHEDULERS.register_module()
class WarmupScheduler(LinearScheduler):

    def __init__(self, *, value: float = 1, iter_: int) -> None:
        super().__init__(
            start_value=0,
            end_value=value,
            start_iter=0,
            end_iter=iter_,
        )


@SCHEDULERS.register_module()
class EarlyStopScheduler(LinearScheduler):

    def __init__(self, *, value: float = 1, iter_: int) -> None:
        super().__init__(
            start_value=value,
            end_value=0,
            start_iter=iter_,
            end_iter=iter_,
        )


@SCHEDULERS.register_module()
class DecayScheduler(LinearScheduler):

    def __init__(self, *, value: float = 1, iter_: int) -> None:
        super().__init__(
            start_value=value,
            end_value=0,
            start_iter=0,
            end_iter=iter_,
        )



import einops
import torch
import torch.nn as nn

from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class Decouple(BaseAdapt):

    def __init__(
        self,
        num: int,
        in_features: int,
        out_features: int,
        bias: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._num = num
        self._layer = nn.Linear(in_features, out_features * num, bias)

    def forward(self, feat: torch.Tensor, id_: torch.Tensor) -> torch.Tensor:
        """Decouple `feat`.

        Args:
            feat: n x dim
            pos: n

        Returns:
            decoupled_feat: n x dim
        """
        decoupled_feat: torch.Tensor = self._layer(feat)  # n x (num x dim)
        decoupled_feat = einops.rearrange(
            decoupled_feat,
            'n (num dim) -> n num dim',
            num=self._num,
        )
        decoupled_feat = decoupled_feat[  # yapf: disable
            torch.arange(id_.shape[0]), id_.long(),
        ]  # n x dim
        return decoupled_feat



"""Toolkit for Object Detection Distillation."""
from . import (
    adapts,
    distillers,
    hooks,
    losses,
    reproduction,
    schedulers,
    utils,
    visuals,
)
from .base import *

__all__ = [
    'adapts',
    'distillers',
    'hooks',
    'losses',
    'reproduction',
    'schedulers',
    'utils',
    'visuals',
]



import functools
import sys

from .._extensions import get_logger

if sys.version_info < (3, 8):
    get_logger().warning("Monkey patching `functools.cached_property`.")

    class cached_property:

        def __init__(self, func):
            self.func = func
            self.__doc__ = func.__doc__

        def __set_name__(self, owner, name):
            self.attrname = name

        def __get__(self, instance, owner=None):
            if self.attrname in instance.__dict__:
                return instance.__dict__[self.attrname]
            val = self.func(instance)
            instance.__dict__[self.attrname] = val
            return val

    functools.cached_property = cached_property

if sys.version_info < (3, 9):
    get_logger().warning("Monkey patching `functools.cache`.")
    functools.cache = functools.lru_cache(maxsize=None)



import torch

from ..utils import ListTensor
from .base import ADAPTS, BaseAdapt


class ListTensorAdapt(BaseAdapt):
    func: staticmethod

    def forward(self, *args, **kwargs) -> torch.Tensor:
        return self.func(*args, **kwargs)


@ADAPTS.register_module()
class Stack(ListTensorAdapt):
    func = staticmethod(ListTensor.stack)


@ADAPTS.register_module()
class Index(ListTensorAdapt):
    func = staticmethod(ListTensor.index)



import torch

from .._extensions import get_logger

if torch.__version__ < '1.7.0':
    get_logger().warning(
        "Monkey patching `torch.maximum` and `torch.minimum`.",
    )
    torch.maximum = torch.max
    torch.Tensor.maximum = torch.Tensor.max
    torch.minimum = torch.min
    torch.Tensor.minimum = torch.Tensor.min



__all__ = [
    'getattr_recur',
    'hasattr_recur',
    'setattr_recur',
    'delattr_recur',
    'setattr_temp',
]

import contextlib
import functools
from typing import Any, Generator


def patch(patched_func):

    def patcher(func):

        @functools.wraps(func)
        def patch_func(obj, attr, *args, **kwargs):
            if '.' not in attr:
                func = patched_func
            return func(obj, attr, *args, **kwargs)

        return patch_func

    return patcher


@patch(getattr)
def getattr_recur(obj: Any, attr: str, default=...):
    try:
        return eval('obj' + attr)
    except Exception:
        if default is not ...:
            return default
        raise


@patch(hasattr)
def hasattr_recur(obj: Any, attr: str) -> bool:
    default = object()
    return getattr_recur(obj, attr, default) is not default


@patch(setattr)
def setattr_recur(obj, attr: str, value) -> None:
    attr, name = attr.rsplit('.', 1)
    if attr:
        obj = getattr_recur(obj, attr)
    setattr(obj, name, value)


@patch(delattr)
def delattr_recur(obj, attr: str) -> None:
    attr, name = attr.rsplit('.', 1)
    if attr:
        obj = getattr_recur(obj, attr)
    delattr(obj, name)


@contextlib.contextmanager
def setattr_temp(obj, attr: str, value) -> Generator[None, None, None]:
    if hasattr_recur(obj, attr):
        prev = getattr_recur(obj, attr)
        setattr_recur(obj, attr, value)
        yield
        setattr_recur(obj, attr, prev)
    else:
        setattr_recur(obj, attr, value)
        yield
        delattr_recur(obj, attr)



import string

from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class Custom(BaseAdapt):

    def __init__(self, *args, pattern: str, **kwargs):
        super().__init__(*args, **kwargs)
        self._pattern = pattern

    def forward(self, *args, **kwargs):
        for i, (name, _) in enumerate(zip(string.ascii_letters, args)):
            exec(f'{name} = args[{i}]')
        for key in kwargs:
            if key in locals():
                raise SyntaxError(key)
            exec(f'{key} = kwargs["{key}"]')
        return eval(self._pattern)



__all__ = [
    'StandardHook',
]

from .base import HOOKS, BaseHook


@HOOKS.register_module()
class StandardHook(BaseHook):

    def _reset(self):
        self.__tensor = None

    def _tensor(self):
        return self.__tensor

    def _register_tensor(self, tensor) -> None:
        self.__tensor = tensor



__all__ = [
    'SGFILoss',
]

from typing import Callable, List

import einops
import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import LOSSES
from .functional import MSELoss


@LOSSES.register_module()
class SGFILoss(MSELoss):

    def __init__(
        self,
        *args,
        in_channels: int = 256,
        hidden_channels: int = 128,
        out_channels: int = 64,
        **kwargs,
    ):
        from mmcv.cnn import ConvModule

        super().__init__(*args, **kwargs)
        self._embed: Callable[..., torch.Tensor] = nn.Sequential(
            ConvModule(in_channels, hidden_channels, 3, stride=2),
            ConvModule(hidden_channels, out_channels, 3, stride=2)
        )
        self._tau = nn.Parameter(torch.tensor([1.0], dtype=torch.float32))

    def forward(  # type: ignore[override]
        self,
        preds: List[torch.Tensor],
        targets: torch.Tensor,
        *args,
        **kwargs,
    ):
        """Re-implementation of G-DetKD.

        Refer to http://arxiv.org/abs/2108.07482.

        Args:
            pred: l x r x c x h x w
                Each of the `l` levels generate `r` RoIs.
                Typical shape is 4 x 1024 x 256 x 7 x 7.

            target: r x c x h x w

        Returns:
            loss: 1
        """
        fused_preds = torch.stack(preds)
        embed_pred = einops.rearrange(fused_preds, 'l r c h w -> (l r) c h w')
        embed_pred = self._embed(embed_pred)
        embed_target = self._embed(targets)

        embed_pred = einops.rearrange(
            embed_pred,
            '(l r) out_channels 1 1 -> l r out_channels',
            l=len(preds),
        )
        embed_target = einops.rearrange(
            embed_target,
            'r out_channels 1 1 -> r out_channels',
        )
        similarity = torch.einsum(
            'l r c, r c -> l r',
            embed_pred,
            embed_target,
        )
        similarity = F.softmax(similarity / self._tau, dim=1)

        fused_pred = torch.einsum(
            'l r c h w, l r -> r c h w',
            fused_preds,
            similarity,
        )
        return super().forward(fused_pred, targets, *args, **kwargs)


# @LOSSES.register_module()
# class DevRCNNLoss(MSELoss):
#     def __init__(
#         self,
#         *args,
#         pred_features: int = 256,
#         target_features: int = 1024,
#         **kwargs,
#     ):
#         super().__init__(*args, **kwargs)
#         self._adapt: Callable[..., torch.Tensor] = nn.Linear(
#             in_features=pred_features,
#             out_features=target_features,
#             bias=False,
#         )
#         self._tau = nn.Parameter(torch.FloatTensor(data=[1]))

#     def forward(
#         self, preds: List[torch.Tensor], targets: torch.Tensor,
#         poses: torch.Tensor,
#         *args, **kwargs,
#     ):
#         """
#         Args:
#             pred: l x bs x h x w x in_features
#             target: r x c
#             pos: r x 4

#         Returns:
#             loss: 1
#         """
#         l = len(preds)
#         poses[:, 2:] *= 2 ** poses[:, [0]]
#         poses[:, 0] = 0
#         for i in range(l):
#             preds[i] = ListTensor.index(preds[i], poses[:, 1:])
#             poses[:, 2:] = poses[:, 2:] // 2
#         preds = einops.rearrange(
#             torch.stack(preds),
#             'l r pred_features -> (l r) pred_features',
#         )
#         preds = self._adapt(preds)
#         preds = einops.rearrange(
#             preds,
#             '(l r) target_features -> r l target_features',
#             l=l,
#         )
#         targets = einops.rearrange(
#             targets,
#             'r target_features -> r target_features 1',
#         )

#         similarity = preds.bmm(targets)  # r x l x 1
#         similarity = F.softmax(similarity / self._tau, dim=1)

#         fused_pred = einops.reduce(
#             preds * similarity,
#             'r l target_features -> r target_features 1',
#             reduction='sum',
#         )
#         return super().forward(fused_pred, targets, *args, **kwargs)



import math
import numbers
from abc import abstractmethod
from typing import Any, Optional, Union

from ..base import Registry, get_iter

__all__ = [
    'BaseScheduler',
    'SCHEDULERS',
    'SchedulerConfig',
    'IntervalScheduler',
]


class BaseScheduler(numbers.Real):

    @property
    @abstractmethod
    def value(self) -> float:
        pass

    def __add__(self, other):
        return self.value + other

    def __radd__(self, other):
        return other + self.value

    def __neg__(self):
        return -self.value

    def __pos__(self):
        return +self.value

    def __mul__(self, other):
        return self.value * other

    def __rmul__(self, other):
        return other * self.value

    def __truediv__(self, other):
        return self.value / other

    def __rtruediv__(self, other):
        return other / self.value

    def __pow__(self, exponent):
        return self.value**exponent

    def __rpow__(self, base):
        return base**self.value

    def __abs__(self):
        return abs(self.value)

    def __eq__(self, other):
        return self.value == other

    def __float__(self) -> float:
        return float(self.value)

    def __trunc__(self) -> int:
        return math.trunc(self.value)

    def __floor__(self) -> int:
        return math.floor(self.value)

    def __ceil__(self) -> int:
        return math.ceil(self.value)

    def __round__(self, ndigits=None):
        return round(self.value, ndigits)

    def __floordiv__(self, other):
        return self.value // other

    def __rfloordiv__(self, other):
        return other // self.value

    def __mod__(self, other):
        return self.value % other

    def __rmod__(self, other):
        return other % self.value

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other


SCHEDULERS: Registry[Any] = Registry(
    'schedulers',
    base=BaseScheduler,
)
SchedulerConfig = Union[dict, BaseScheduler]


class IntervalScheduler(BaseScheduler):

    def __init__(
        self,
        *,
        start_value: float,
        end_value: float,
        start_iter: int = 0,
        end_iter: Optional[int] = None,
    ) -> None:
        super().__init__()
        self._start_value = start_value
        self._end_value = end_value
        self._start_iter = start_iter
        self._end_iter = float('inf') if end_iter is None else end_iter

    @abstractmethod
    def _weight(
        self,
        cur_iter: int,
        total_iter: float,  # may be float('inf')
    ) -> float:
        pass

    @property
    def value(self) -> float:
        if get_iter() <= self._start_iter:
            return self._start_value
        if get_iter() >= self._end_iter:
            return self._end_value
        weight = self._weight(
            get_iter() - self._start_iter,
            self._end_iter - self._start_iter,
        )
        return (
            weight * (self._end_value - self._start_value) + self._start_value
        )



import argparse
import re
from enum import Enum, auto
from typing import Iterable, Mapping, MutableMapping

import torch
from mmcv import DictAction


class MatchMode(Enum):
    PREFIX = auto()
    REGEX = auto()
    EXACT = auto()

    def match(self, key: str, pattern: str) -> bool:
        if self == MatchMode.PREFIX:
            return key.startswith(pattern)
        if self == MatchMode.REGEX:
            return re.match(pattern, key) is not None
        if self == MatchMode.EXACT:
            return key == pattern
        raise ValueError(f"Unknown match mode: {self}")


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Convert Checkpoints")
    parser.add_argument('ckpt')
    for mode in MatchMode:
        parser.add_argument(
            f'--rename-{mode.name.lower()}',
            nargs='+',
            action=DictAction,
        )
        parser.add_argument(f'--remove-{mode.name.lower()}', nargs='+')
    parser.add_argument('--out')
    args = parser.parse_args()
    return args


def rename_exact(
    state_dict: MutableMapping[str, torch.Tensor],
    mapping: Mapping[str, str],
):
    for k, v in mapping.items():
        state_dict[v] = state_dict[k]
        del state_dict[k]


def rename(
    state_dict: MutableMapping[str, torch.Tensor],
    mapping: Mapping[str, str],
    mode: MatchMode,
):
    if mode == MatchMode.EXACT:
        rename_exact(state_dict, mapping)
        return
    exact_mapping = {}
    for k, v in state_dict.items():
        patterns = {pattern for pattern in mapping if mode.match(k, pattern)}
        if len(patterns) == 0:
            continue
        longest_pattern = max(patterns, key=len)
        exact_mapping[k] = k.replace(longest_pattern, mapping[longest_pattern])
    rename_exact(state_dict, exact_mapping)


def remove_exact(
    state_dict: MutableMapping[str, torch.Tensor],
    patterns: Iterable[str],
):
    for k in patterns:
        del state_dict[k]


def remove(
    state_dict: MutableMapping[str, torch.Tensor],
    patterns: Iterable[str],
    mode: MatchMode,
):
    if mode == MatchMode.EXACT:
        remove_exact(state_dict, patterns)
        return
    exact_patterns = set()
    for k in state_dict:
        if any(mode.match(k, pattern) for pattern in patterns):
            exact_patterns.add(k)
    remove_exact(state_dict, exact_patterns)


def main():
    args = parse_args()

    ckpt = torch.load(args.ckpt, 'cpu')
    state_dict = ckpt
    if 'state_dict' in state_dict:
        state_dict = state_dict['state_dict']

    if args.out is None:
        args.out = args.ckpt + '.converted'

    for mode in MatchMode:
        mapping = getattr(args, f'rename_{mode.name.lower()}')
        if mapping is not None:
            rename(state_dict, mapping, mode)

    for mode in MatchMode:
        patterns = getattr(args, f'remove_{mode.name.lower()}')
        if patterns is not None:
            remove(state_dict, patterns, mode)

    torch.save(state_dict, args.out)


if __name__ == '__main__':
    main()



from abc import abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Generic, Optional, Sequence, TypeVar

import cv2
import numpy as np

from ..base import get_iter
from .base import VISUALS, BaseVisual

__all__ = [
    'BaseSaver',
    'CV2Saver',
]

T = TypeVar('T')


class BaseSaver(Generic[T], BaseVisual):

    def __init__(
        self,
        *args,
        root_dir: str,
        trial_name: Optional[str] = None,
        suffix: str = '',
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._root_dir = Path(root_dir)
        self._trial_name = (
            datetime.now().strftime("%Y%m%dT%H%M%S%f")
            if trial_name is None else trial_name
        )
        self._suffix = suffix

        visual_dir = self._root_dir / self._trial_name
        if not visual_dir.exists():
            visual_dir.mkdir(parents=True)
        self._visual_dir = visual_dir

    def _get_path(self, **kwargs) -> Path:
        assert 'iter' not in kwargs
        kwargs['iter'] = get_iter()
        filename = '_'.join(f'{k}-{v}' for k, v in kwargs.items())
        path = self._visual_dir / f'{filename}_{self._suffix}'
        return path

    @abstractmethod
    def forward(self, data: Sequence[T], **kwargs) -> None:
        pass


@VISUALS.register_module()
class CV2Saver(BaseSaver[np.ndarray]):

    def __init__(self, *args, suffix: str = '', **kwargs):
        super().__init__(*args, suffix=suffix + '.png', **kwargs)

    def forward(self, data: Sequence[np.ndarray], **kwargs) -> None:
        for sample, image in enumerate(data):
            path = self._get_path(sample=sample, **kwargs)
            result = cv2.imwrite(str(path), image)
            assert result, path



import einops
import torch
import torch.nn.functional as F

from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class AbsMeanSpatialAttention(BaseAdapt):

    def __init__(self, *args, temperature: float = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self._temperature = temperature

    def forward(self, feat: torch.Tensor) -> torch.Tensor:
        _, _, h, w = feat.shape
        g = einops.reduce(
            torch.abs(feat),
            'n c h w -> n (h w)',
            reduction='mean',
        )
        a = h * w * F.softmax(g / self._temperature, dim=1)
        return einops.rearrange(a, 'n (h w) -> n 1 h w', h=h, w=w)


@ADAPTS.register_module()
class AbsMeanChannelAttention(BaseAdapt):

    def __init__(self, *args, temperature: float = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self._temperature = temperature

    def forward(self, feat: torch.Tensor) -> torch.Tensor:
        _, c, _, _ = feat.shape
        g = einops.reduce(torch.abs(feat), 'n c h w -> n c', reduction='mean')
        a = c * F.softmax(g / self._temperature, dim=1)
        return einops.rearrange(a, 'n c -> n c 1 1')



import itertools
from enum import Enum, auto
from typing import Optional

import torch.nn as nn
from timm.models.layers import DropPath
from torch.nn.modules.batchnorm import _BatchNorm

from ..base import getattr_recur


class NoGradMode(Enum):
    ALL = auto()
    PARTIAL = auto()
    NONE = auto()

    def no_grad(self, model: nn.Module, **kwargs) -> nn.Module:
        if self == NoGradMode.ALL:
            return model.requires_grad_(False)
        elif self == NoGradMode.PARTIAL:
            for module_name in kwargs['modules']:
                module: nn.Module = getattr_recur(model, module_name)
                module.requires_grad_(False)
            return model
        elif self == NoGradMode.NONE:
            return model
        raise NotImplementedError(f"{self} is not implemented")


DEFAULT_STOCHASTIC_MODULE_TYPES = [_BatchNorm, nn.Dropout, DropPath]


class EvalMode(Enum):
    ALL = auto()
    DETERMINISTIC_AND_PARTIAL = auto()
    DETERMINISTIC = auto()
    PARTIAL = auto()
    PARTIALLY_DETERMINISTIC = auto()
    NONE = auto()

    def eval(self, model: nn.Module, **kwargs) -> nn.Module:
        if self == EvalMode.ALL:
            return model.eval()
        elif self == EvalMode.NONE:
            return model
        # TODO: refactor
        modules = (
            model.modules() if self == EvalMode.DETERMINISTIC else
            (getattr_recur(model, module) for module in kwargs['modules'])
        )
        if self != EvalMode.PARTIAL:
            sms = kwargs.get(
                'stochastic_module_types',
                DEFAULT_STOCHASTIC_MODULE_TYPES,
            )
            if self in [
                EvalMode.DETERMINISTIC,
                EvalMode.PARTIALLY_DETERMINISTIC,
            ]:
                modules = (
                    module for module in modules
                    if any(isinstance(module, sm) for sm in sms)
                )
            elif self == EvalMode.DETERMINISTIC_AND_PARTIAL:
                modules = itertools.chain(
                    modules,
                    (
                        module for module in model.modules()
                        if any(isinstance(module, sm) for sm in sms)
                    ),
                )
            else:
                raise NotImplementedError(f'EvalMode {self} not implemented')
        for module in modules:
            module.eval()
        return model


def freeze_model(
    model: nn.Module,
    no_grad: Optional[dict] = ...,  # type: ignore[assignment]
    eval_: Optional[dict] = ...,  # type: ignore[assignment]
) -> nn.Module:
    if no_grad is ...:
        no_grad = dict(mode='ALL')
    if no_grad is not None:
        no_grad = dict(no_grad)
        no_grad_mode_name: str = no_grad.pop('mode')
        no_grad_mode = NoGradMode[no_grad_mode_name.upper()]
        no_grad_mode.no_grad(model, **no_grad)
    if eval_ is ...:
        eval_ = dict(mode='ALL')
    if eval_ is not None:
        eval_ = dict(eval_)
        eval_mode_name: str = eval_.pop('mode')
        eval_mode = EvalMode[eval_mode_name.upper()]
        eval_mode.eval(model, **eval_)
    return model


class FrozenMixin(nn.Module):

    def __init__(
        self,
        *args,
        freeze_cfg: Optional[dict] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if freeze_cfg is None:
            return
        freeze_model(self, **freeze_cfg)
        self._eval_cfg = freeze_cfg.get('eval_')

    def train(self, mode: bool = True):
        result = super().train(mode)
        if hasattr(self, '_eval_cfg'):
            freeze_model(self, no_grad=None, eval_=self._eval_cfg)
        return result



__all__ = [
    'MultiCallsHook',
]

from typing import List

import torch

from .base import HOOKS, BaseHook


@HOOKS.register_module()
class MultiCallsHook(BaseHook):

    def _reset(self):
        self._tensors = []

    def _tensor(self) -> List[torch.Tensor]:
        return self._tensors

    def _register_tensor(self, tensor: torch.Tensor) -> None:
        self._tensors.append(tensor)



from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..base import Module, Sequential
from .base import ADAPTS, BaseAdapt


class Bottleneck(Module):

    def __init__(
        self,
        in_channels,
        hidden_channels,
        out_channels,
        downsample: bool = False,
    ):
        super().__init__()
        stride = 2 if downsample else 1
        self._downsample = None if in_channels == out_channels else nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=1,
            stride=stride,
            bias=False,
        )
        self._conv1 = nn.Conv2d(
            in_channels,
            hidden_channels,
            kernel_size=1,
            stride=1,
        )
        self._conv2 = nn.Conv2d(
            hidden_channels,
            hidden_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
        )
        self._conv3 = nn.Conv2d(
            hidden_channels,
            out_channels,
            kernel_size=1,
        )

    def init_weights(self):
        assert not self.is_init
        for layer in self.children():
            assert isinstance(layer, nn.Conv2d)
            nn.init.kaiming_uniform_(layer.weight, a=1)
        self._is_init = True

    def forward(self, x):
        identity = x

        out = self._conv1(x)
        out = F.relu_(out)
        out = self._conv2(out)
        out = F.relu_(out)
        out = self._conv3(out)

        if self._downsample is not None:
            identity = self._downsample(x)

        out += identity
        out = F.relu_(out)
        return out


class ResBlock(Sequential):

    def __init__(
        self,
        *,
        base_channels: int,
        expansion: int = 2,
        **kwargs,
    ):
        layer1 = Bottleneck(
            base_channels * expansion,
            base_channels,
            base_channels * expansion**2,
            downsample=True,
        )
        layer2 = Bottleneck(
            base_channels * expansion**2,
            base_channels,
            base_channels * expansion**2,
        )
        super().__init__(layer1, layer2, **kwargs)


@ADAPTS.register_module()
class LabelEncAdapt(BaseAdapt):

    def __init__(
        self,
        *args,
        num_classes: int = 80,
        base_channels: int = 32,
        expansion: int = 2,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._stage1 = nn.Conv2d(
            num_classes,
            base_channels * 4,
            kernel_size=7,
            stride=2,
            padding=3,
        )
        base_channels *= 2
        self._stage2 = Bottleneck(
            base_channels * 2,
            base_channels,
            base_channels * 4,
            downsample=True,
        )
        base_channels *= 2
        self._stage3 = ResBlock(
            base_channels=base_channels,
            expansion=expansion,
        )
        base_channels *= 2
        self._stage4 = ResBlock(
            base_channels=base_channels,
            expansion=expansion,
        )
        base_channels *= 2
        self._stage5 = Bottleneck(
            base_channels * 2,
            base_channels,
            base_channels * 4,
            downsample=True,
        )

    def forward(self, x: torch.Tensor) -> List[torch.Tensor]:
        x1 = self._stage1(x)
        x2 = self._stage2(x1)
        x3 = self._stage3(x2)
        x4 = self._stage4(x3)
        x5 = self._stage5(x4)
        outs = [x2, x3, x4, x5]
        return outs



import torch.nn as nn

from ..base import STEPS, Module, Registry

__all__ = [
    'BaseVisual',
    'VISUALS',
]


class BaseVisual(Module):
    pass


VISUALS: Registry[nn.Module] = Registry(
    'visuals',
    parent=STEPS,
    base=nn.Module,
)



from typing import List, Tuple

import torch

from ..base import BBoxesXYXY
from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class IoU(BaseAdapt):

    def __init__(
        self,
        *args,
        aligned: bool = False,
        eps: float = 1e-6,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert not aligned
        self._aligned = aligned
        self._eps = eps

    def _reshape(
        self,
        bboxes: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Size]:
        """
        Args:
            bboxes: * x 4

        Returns:
            bboxes: prod(*) x 4
            shape: tuple(*)
        """
        bboxes = bboxes[..., :4]
        # bboxes = bboxes.half()
        shape = bboxes.shape[:-1]
        bboxes = bboxes.reshape(-1, 4)
        return bboxes, shape

    def _iou(
        self,
        bboxes1: torch.Tensor,
        bboxes2: torch.Tensor,
    ) -> torch.Tensor:
        """
        Args:
            bboxes1: *1 x 4
            bboxes2: *2 x 4

        Returns:
            ious: *1 x *2
        """
        bboxes1, shape1 = self._reshape(bboxes1)
        bboxes2, shape2 = self._reshape(bboxes2)
        ious = BBoxesXYXY.ious(
            BBoxesXYXY(bboxes1),
            BBoxesXYXY(bboxes2),
            self._eps,
        )
        return ious.reshape(shape1 + shape2)

    def forward(
        self,
        bboxes1: List[torch.Tensor],
        bboxes2: List[torch.Tensor],
    ) -> List[List[torch.Tensor]]:
        """
        Args:
            bboxes1: n1 x m1 x 4
            bboxes2: n2 x m2 x 4

        Returns:
            ious: n1 x n2 x m1 x m2
        """
        return [[self._iou(bbox1, bbox2)
                 for bbox2 in bboxes2]
                for bbox1 in bboxes1]



__all__ = [
    'BBoxesXYXY',
    'BBoxesXYWH',
]

from abc import abstractmethod
from typing import (
    List,
    NamedTuple,
    Optional,
    Sized,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import einops
import numpy as np
import torch

T = TypeVar('T', bound='BBoxes')


class BBoxes(Sized):

    def __init__(self, bboxes: Union[np.ndarray, torch.Tensor]) -> None:
        """
        Args:
            bboxes: \\* x 4
                (x1, y1, x2, y2)
        """
        if isinstance(bboxes, np.ndarray):
            bboxes = torch.from_numpy(bboxes)
        if bboxes.numel() == 0:
            raise ValueError('bboxes is empty')
        if bboxes.ndim < 2:
            raise ValueError('bboxes must be at least 2-dim')
        if bboxes.shape[-1] != 4:
            raise ValueError('bboxes must have 4 columns')
        self._bboxes = bboxes

    def __len__(self) -> int:
        return self._bboxes.shape[0]

    @property
    @abstractmethod
    def left(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def right(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def top(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def bottom(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def width(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def height(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def center_x(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def center_y(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def lt(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def rb(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def wh(self) -> torch.Tensor:
        pass

    @property
    @abstractmethod
    def center(self) -> torch.Tensor:
        pass

    def to_tensor(self) -> torch.Tensor:
        return self._bboxes

    @property
    def shapes(self) -> torch.Tensor:
        return self.wh

    @property
    def areas(self) -> torch.Tensor:
        return self.width * self.height

    @classmethod
    def _cat(cls: Type[T], a: T, b: T) -> T:
        """
        Args:
            a: *1 x 4
            b: *2 x 4

        Returns:
            ious: (*1 + *2) x 4
        """
        bboxes = torch.cat([a.to_tensor(), b.to_tensor()])
        return cls(bboxes)

    def cat(self: T, other: T) -> T:
        return self._cat(self, other)

    def __add__(self: T, other: T) -> T:
        return self.cat(other)

    @classmethod
    def _select(
        cls: Type[T],
        bboxes: T,
        indices: Union[int, slice, torch.Tensor, List[int], Tuple[int, ...]],
    ) -> T:
        """
        Args:
            bboxes: n x 4
            indices: m

        Returns:
            bboxes: m x 4
        """
        return cls(bboxes.to_tensor()[indices])

    def select(
        self: T,
        indices: Union[int, slice, torch.Tensor, List[int], Tuple[int, ...]],
    ) -> T:
        return self._select(self, indices)

    def __getitem__(
        self: T,
        indices: Union[int, slice, torch.Tensor, List[int], Tuple[int, ...]],
    ) -> T:
        return self.select(indices)

    @classmethod
    def _intersections(cls: Type[T], a: T, b: T) -> torch.Tensor:
        """
        Args:
            a: *1 x 4
            b: *2 x 4

        Returns:
            intersections: *1 x *2
        """
        lt = torch.maximum(  # [*1, *2, 2]
            einops.rearrange(a.lt, 'n1 lt -> n1 1 lt'),
            einops.rearrange(b.lt, 'n2 lt -> 1 n2 lt'),
        )
        rb = torch.minimum(  # [*1, *2, 2]
            einops.rearrange(a.rb, 'n1 rb -> n1 1 rb'),
            einops.rearrange(b.rb, 'n2 rb -> 1 n2 rb'),
        )
        wh = rb - lt
        wh = wh.clamp_min_(0)
        return wh[..., 0] * wh[..., 1]

    def intersections(self: T, other: T) -> torch.Tensor:
        return self._intersections(self, other)

    def __and__(self: T, other: T) -> torch.Tensor:
        return self.intersections(other)

    @classmethod
    def _unions(
        cls: Type[T],
        a: T,
        b: T,
        intersections: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Args:
            a: *1 x 4
            b: *2 x 4
            intersections: *1 x *2

        Returns:
            unions: *1 x *2
        """
        if intersections is None:
            intersections = cls.intersections(a, b)
        return a.areas[:, None] + b.areas[None, :] - intersections

    def unions(
        self: T,
        other: T,
        intersections: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        return self._unions(self, other, intersections)

    def __or__(self: T, other: T) -> torch.Tensor:
        return self.unions(other)

    def ious(self: T, other: T, eps: float = 1e-6) -> torch.Tensor:
        """
        Args:
            a: *1 x 4
            b: *2 x 4

        Returns:
            ious: *1 x *2
        """
        intersections = self & other
        unions = self.unions(other, intersections).clamp_min_(eps)
        ious = intersections / unions
        return ious

    @classmethod
    @abstractmethod
    def _round(cls: Type[T], bboxes: T) -> T:
        pass

    def round(self: T) -> T:
        return self._round(self)

    @classmethod
    @abstractmethod
    def _expand(
        cls: Type[T],
        bboxes: T,
        ratio: float,
        image_shape: Optional[Tuple[int, int]] = None,
    ) -> T:
        """
        Args:
            bboxes: n x 4

        Returns:
            expanded_bboxes: n x 4
        """
        pass

    def expand(
        self: T,
        ratio: float,
        image_shape: Optional[Tuple[int, int]] = None,
    ) -> T:
        return self._expand(self, ratio, image_shape)


T_XY = TypeVar('T_XY', bound='BBoxesXY')


class BBoxesXY(BBoxes):
    _BBoxType = NamedTuple(
        '_BBoxType',
        [('lt', torch.Tensor), ('rb', torch.Tensor)],
    )

    @property
    def left(self) -> torch.Tensor:
        return self._bboxes[:, 0]

    @property
    def top(self) -> torch.Tensor:
        return self._bboxes[:, 1]

    @property
    def lt(self) -> torch.Tensor:
        return self._bboxes[:, :2]

    def _round_lt_rb(self) -> 'BBoxesXY._BBoxType':
        lt = self.lt.floor()
        rb = self.rb.ceil()
        return self._BBoxType(lt, rb)

    def _expand_lt_rb(
        self,
        ratio: float,
        image_shape: Optional[Tuple[int, int]] = None,
    ) -> 'BBoxesXY._BBoxType':
        offsets = self.wh * (ratio - 1) / 2
        lt = self.lt - offsets
        rb = self.rb + offsets

        lt.clamp_min_(0)
        if image_shape is not None:
            h, w = image_shape
            rb[:, 0].clamp_max_(w)
            rb[:, 1].clamp_max_(h)

        return self._BBoxType(lt, rb)


class BBoxesXYXY(BBoxesXY):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert self.wh.gt(0).all()

    @property
    def right(self) -> torch.Tensor:
        return self._bboxes[:, 2]

    @property
    def bottom(self) -> torch.Tensor:
        return self._bboxes[:, 3]

    @property
    def width(self) -> torch.Tensor:
        return self.right - self.left

    @property
    def height(self) -> torch.Tensor:
        return self.bottom - self.top

    @property
    def center_x(self) -> torch.Tensor:
        return (self.left + self.right) / 2

    @property
    def center_y(self) -> torch.Tensor:
        return (self.top + self.bottom) / 2

    @property
    def rb(self) -> torch.Tensor:
        return self._bboxes[:, 2:]

    @property
    def wh(self) -> torch.Tensor:
        return self.rb - self.lt

    @property
    def center(self) -> torch.Tensor:
        return (self.lt + self.rb) / 2

    @classmethod
    def _round(cls: Type[T_XY], bboxes: T_XY) -> T_XY:
        lt, rb = bboxes._round_lt_rb()
        rounded_bboxes = torch.cat([lt, rb], dim=-1)
        return cls(rounded_bboxes)

    @classmethod
    def _expand(
        cls: Type[T_XY],
        bboxes: T_XY,
        ratio: float,
        image_shape: Optional[Tuple[int, int]] = None,
    ) -> T_XY:
        lt, rb = bboxes._expand_lt_rb(ratio, image_shape)
        expanded_bboxes = torch.cat([lt, rb], dim=-1)
        return cls(expanded_bboxes)


class BBoxesXYWH(BBoxesXY):

    @property
    def width(self) -> torch.Tensor:
        return self._bboxes[:, 2]

    @property
    def height(self) -> torch.Tensor:
        return self._bboxes[:, 3]

    @property
    def right(self) -> torch.Tensor:
        return self.left + self.width

    @property
    def bottom(self) -> torch.Tensor:
        return self.top + self.height

    @property
    def center_x(self) -> torch.Tensor:
        return self.left + self.width / 2

    @property
    def center_y(self) -> torch.Tensor:
        return self.top + self.height / 2

    @property
    def wh(self) -> torch.Tensor:
        return self._bboxes[:, 2:]

    @property
    def rb(self) -> torch.Tensor:
        return self.lt + self.wh

    @property
    def center(self) -> torch.Tensor:
        return self.lt + self.wh / 2

    @classmethod
    def _round(cls: Type[T_XY], bboxes: T_XY) -> T_XY:
        lt, rb = bboxes._round_lt_rb()
        rounded_bboxes = torch.cat([lt, rb - lt], dim=-1)
        return cls(rounded_bboxes)

    @classmethod
    def _expand(
        cls: Type[T_XY],
        bboxes: T_XY,
        ratio: float,
        image_shape: Optional[Tuple[int, int]] = None,
    ) -> T_XY:
        lt, rb = bboxes._expand_lt_rb(ratio, image_shape)
        expanded_bboxes = torch.cat([lt, rb - lt], dim=-1)
        return cls(expanded_bboxes)



import functools
import os
import sys
import typing
import zipfile

from .._extensions import get_logger

if sys.version_info < (3, 8):
    get_logger().warning("Monkey patching `zipfile.Path`.")

    class Path:

        def __init__(self, root: str, at: str = ''):
            self._root = zipfile.ZipFile(root)
            self._at = at

        def _next(self, at: str) -> 'Path':
            return Path(self._root.filename, at)

        def _is_child(self, path: 'Path') -> bool:
            return (
                os.path.dirname(path._at.rstrip("/")) == self._at.rstrip("/")
            )

        @functools.cached_property
        def _namelist(self) -> typing.List[str]:
            return self._root.namelist()

        def read_bytes(self) -> bytes:
            return self._root.read(self._at)

        def exists(self) -> bool:
            return self._at in self._namelist

        def iterdir(self) -> typing.Iterator['Path']:
            subs = map(self._next, self._namelist)
            return filter(self._is_child, subs)

        def __str__(self):
            return os.path.join(self._root.filename, self._at)

        def __truediv__(self, at: str) -> 'Path':
            return self._next(os.path.join(self._at, at))

    zipfile.Path = Path



from .attention import AbsMeanChannelAttention, AbsMeanSpatialAttention
from .base import *
from .custom import Custom
from .decouple import Decouple
from .detach import Detach, ListDetach
from .dict_tensor import *
from .iou import IoU
from .label_enc import LabelEncAdapt
from .list_tensor import Index, Stack
from .mask import DeFeatMask, FGDMask, FGFIMask, LabelEncMask
from .null import Null
from .roi_align import RoIAlign

__all__ = [
    'AbsMeanSpatialAttention',
    'AbsMeanChannelAttention',
    'Custom',
    'Decouple',
    'Detach',
    'ListDetach',
    'IoU',
    'LabelEncAdapt',
    'Stack',
    'Index',
    'DeFeatMask',
    'FGDMask',
    'FGFIMask',
    'LabelEncMask',
    'Null',
    'RoIAlign',
]



import numbers
from typing import Callable, Generic, Iterable, Literal, Tuple, TypeVar, Union

import torch

T = TypeVar('T', torch.Tensor, list, tuple, dict)


class CollectionTensor(Generic[T]):

    @staticmethod
    def apply(feat: T, op: Callable[[torch.Tensor], torch.Tensor]) -> T:
        if isinstance(feat, torch.Tensor):
            return op(feat)
        if isinstance(feat, list):
            return [CollectionTensor.apply(f, op) for f in feat]
        if isinstance(feat, tuple):
            return tuple(  # type: ignore[return-value]
                CollectionTensor.apply(f, op) for f in feat
            )
        if isinstance(feat, dict):
            return {k: CollectionTensor.apply(v, op) for k, v in feat.items()}
        raise TypeError(f'Unknown type {type(feat)}.')

    @staticmethod
    def to(feat: T, device: torch.device) -> T:
        return CollectionTensor.apply(feat, lambda x: x.to(device))

    @staticmethod
    def cpu(feat: T) -> T:
        return CollectionTensor.apply(feat, lambda x: x.cpu())

    @staticmethod
    def cuda(feat: T) -> T:
        return CollectionTensor.apply(feat, lambda x: x.cuda())

    @staticmethod
    def allclose(
        feat1: Union[T, numbers.Complex],
        feat2: Union[T, numbers.Complex],
        rtol: float = 1e-5,
        atol: float = 1e-8,
    ) -> bool:
        if isinstance(feat1, torch.Tensor):
            assert isinstance(feat2, torch.Tensor)
            return torch.allclose(feat1, feat2)
        if isinstance(feat1, list) or isinstance(feat1, tuple):
            assert isinstance(feat2, list) or isinstance(feat2, tuple)
            assert len(feat1) == len(feat2)
            return all(  # yapf: disable
                CollectionTensor.allclose(f1, f2)
                for f1, f2 in zip(feat1, feat2)
            )
        if isinstance(feat1, dict):
            assert isinstance(feat2, dict) and feat1.keys() == feat2.keys()
            return all(
                CollectionTensor.allclose(feat1[k], feat2[k]) for k in feat1
            )
        if isinstance(feat1, numbers.Complex):
            assert isinstance(feat2, numbers.Complex)
            return abs(feat1 - feat2) <= atol + rtol * abs(feat2)
        raise TypeError(f'Unknown type {type(feat1)}.')

    @staticmethod
    def reduce(
        feat: T,
        tensor_op: Callable[[torch.Tensor], torch.Tensor],
        tensors_op: Callable[[Iterable[Union[torch.Tensor, Literal[0]]]],
                             Union[torch.Tensor, Literal[0]]],
    ) -> Union[torch.Tensor, Literal[0]]:
        if isinstance(feat, torch.Tensor):
            return tensor_op(feat)
        if isinstance(feat, list) or isinstance(feat, tuple):
            return tensors_op([
                CollectionTensor.reduce(f, tensor_op, tensors_op) for f in feat
            ])
        if isinstance(feat, dict):
            return tensors_op([
                CollectionTensor.reduce(f, tensor_op, tensors_op)
                for f in feat.values()
            ])
        raise TypeError(f'Unknown type {type(feat)}.')

    @staticmethod
    def sum(feat: T) -> Union[torch.Tensor, Literal[0]]:
        return CollectionTensor.reduce(feat, torch.sum, sum)


class ListTensor(CollectionTensor[T]):

    @staticmethod
    def stack(feat: T) -> torch.Tensor:
        if isinstance(feat, torch.Tensor):
            return feat
        return torch.stack([ListTensor.stack(f) for f in feat])

    @staticmethod
    def shape(feat: T, depth: int = 0) -> Tuple[int, ...]:
        if isinstance(feat, torch.Tensor):
            return feat.shape[max(depth, 0):]
        shapes = {ListTensor.shape(f, depth - 1) for f in feat}
        assert len(shapes) == 1
        shape = shapes.pop()
        if depth <= 0:
            shape = (len(feat), ) + shape
        return shape

    @staticmethod
    def new_empty(feat: T, *args, **kwargs) -> torch.Tensor:
        if isinstance(feat, torch.Tensor):
            return feat.new_empty(*args, **kwargs)
        return ListTensor.new_empty(feat[0], *args, **kwargs)

    @staticmethod
    def index(feat: T, pos: torch.Tensor) -> torch.Tensor:
        """Generalized `feat[pos]`.

        Args:
            feat: d_0 x d_1 x ... x d_(n-1) x *
            pos: m x n

        Returns:
            indexed_feat: m x *
        """
        m, n = pos.shape
        if isinstance(feat, torch.Tensor):
            assert n <= feat.ndim
        if m == 0:
            shape = ListTensor.shape(feat, n)
            return ListTensor.new_empty(feat, 0, *shape)
        if n == 0:
            indexed_feat = ListTensor.stack(feat)
            return indexed_feat.unsqueeze(0).repeat(
                m,
                *[1] * indexed_feat.ndim,
            )

        pos = pos.long()
        if isinstance(feat, torch.Tensor):
            assert (pos >= 0).all()
            max_pos = pos.max(0).values
            feat_shape = pos.new_tensor(feat.shape)
            assert (max_pos < feat_shape[:n]).all(), \
                f'max_pos({max_pos}) larger than feat_shape({feat_shape}).'
            return feat[pos.split(1, 1)].squeeze(1)
        indices = []
        indexed_feats = []
        for i, f in enumerate(feat):
            index = pos[:, 0] == i
            if not index.any():
                continue
            index, = torch.where(index)
            indexed_feat = ListTensor.index(f, pos[index, 1:])
            indices.append(index)
            indexed_feats.append(indexed_feat)
        indexed_feat = torch.cat(indexed_feats)
        index = torch.cat(indices)
        assert index.shape == pos.shape[:1]
        indexed_feat[index] = indexed_feat.clone()
        return indexed_feat



__all__ = [
    'Message',
    'STEPS',
    'WorkflowConfig',
    'Workflow',
]

from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    Union,
)

from ._extensions import Config, get_logger
from .registries import Registry

Message = Dict[str, Any]

Step = Callable[..., Any]
STEPS: Registry[Step] = Registry('steps')


class BaseStepManager:

    @abstractmethod
    def __iter__(self) -> Generator[Any, None, None]:
        pass

    @abstractmethod
    def __call__(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        pass

    @property
    @abstractmethod
    def step(self):
        pass


class SingleStepManager(BaseStepManager):

    def __init__(self, step: Step) -> None:
        self._step = step

    def __iter__(self) -> Generator[Any, None, None]:
        yield self._step

    def __call__(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        return self._step(*args, **kwargs)

    @property
    def step(self) -> Step:
        return self._step


class ParallelSingleStepManager(SingleStepManager):

    def __call__(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        super_call = super().__call__
        return tuple(super_call(p_args, kwargs) for p_args in zip(*args))


class ParallelMultiStepManager(BaseStepManager):

    def __init__(self, steps: Iterable[Step]) -> None:
        self._steps = tuple(steps)

    def __iter__(self) -> Generator[Any, None, None]:
        yield from self._steps

    def __call__(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        return tuple(  # yapf: disable
            step(*p_args, **kwargs)
            for step, *p_args in zip(self._steps, *args)
        )

    @property
    def step(self) -> Tuple[Step, ...]:
        return self._steps


class BaseOutputDict:

    @abstractmethod
    def __call__(self, output) -> Dict[str, Any]:
        pass


class SingleOutputDict(BaseOutputDict):

    def __init__(self, key: str) -> None:
        self._key = key

    def __call__(self, data) -> Dict[str, Any]:
        return {self._key: data}


class MultiOutputDict(BaseOutputDict):

    def __init__(self, keys: Iterable[str]) -> None:
        self._keys = tuple(keys)

    def __call__(self, output) -> Dict[str, Any]:
        return dict(zip(self._keys, output))


class ParallelMultiOutputDict(MultiOutputDict):

    def __call__(self, output) -> Dict[str, Tuple[Any, ...]]:
        output = zip(*output)
        return dict(zip(self._keys, output))


OutputKey = Union[str, Tuple[str, ...]]


class Job:

    @staticmethod
    def build_step_manager(
        config: Config,
        registry: Registry[Step],
        parallel,
    ) -> BaseStepManager:
        if isinstance(parallel, bool):
            step = registry.build(config)
            if parallel:
                return ParallelSingleStepManager(step)
            return SingleStepManager(step)
        if isinstance(parallel, int):
            steps = tuple(registry.build(config) for _ in range(parallel))
            return ParallelMultiStepManager(steps)
        if isinstance(parallel, Iterable):
            steps = tuple(
                registry.build(p_config, config) for p_config in parallel
            )
            return ParallelMultiStepManager(steps)
        raise TypeError(
            "`parallel` must be a bool, int, or Iterable, "
            f"but got {type(parallel)}"
        )

    @staticmethod
    def build_output_dict(key: OutputKey, parallel) -> BaseOutputDict:
        if isinstance(key, str):
            return SingleOutputDict(key)
        if parallel:
            return ParallelMultiOutputDict(key)
        return MultiOutputDict(key)

    @staticmethod
    def build(
        step_descendent_name: str,
        config: Config,
        output_key: OutputKey,
    ) -> 'Job':
        registry = STEPS.descendent(step_descendent_name)
        parallel = config.pop('parallel', False)
        fields = config.pop('fields', tuple())
        step_manager = Job.build_step_manager(config, registry, parallel)
        output_dict = Job.build_output_dict(output_key, parallel)
        return Job(step_manager, output_dict, fields)

    def __init__(
        self,
        step_manager: BaseStepManager,
        output_dict: BaseOutputDict,
        fields: Iterable[str] = tuple(),
    ) -> None:
        self._step_manager = step_manager
        self._output_dict = output_dict
        self._fields = tuple(fields)
        self._logger = get_logger()

    def __iter__(self) -> Iterator[Any]:
        return iter(self._step_manager)

    def __call__(self, message: Message, **kwargs) -> Message:
        inputs = tuple(message[field] for field in self._fields)
        try:
            output = self._step_manager(inputs, kwargs)
        except Exception:
            self._logger.error(f"Failed to forward {self}")
            raise
        return self._output_dict(output)

    @property
    def step(self):
        return self._step_manager.step


WorkflowConfig = Dict[OutputKey, Config]


class Workflow:

    @staticmethod
    def build(
        step_descendent_name: str,
        configs: Optional[WorkflowConfig] = None,
        **kwargs: Config,
    ) -> 'Workflow':
        jobs = tuple(  # yapf: disable
            Job.build(step_descendent_name, config, output_key)
            for output_key, config in (configs or kwargs).items()
        )
        return Workflow(jobs)

    def __init__(self, jobs: Iterable[Job]) -> None:
        self._jobs = tuple(jobs)

    def __iter__(self) -> Iterator[Job]:
        return iter(self._jobs)

    def __call__(self, message: Message) -> Message:
        updated: Message = dict()
        for job in self._jobs:
            updates = job(message)
            message.update(updates)
            updated.update(updates)
        return updated



__all__ = [
    'NORM_LAYERS',
    'Registry',
]

import logging
from functools import partial
from typing import (
    Callable,
    Dict,
    Iterable,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
)

import torch.nn as nn

from ._extensions import Config, get_logger

T = TypeVar('T')


class _RegistryProto(Protocol[T]):  # TODO: inherit from MutatbleMapping
    _name: str
    _logger: logging.Logger
    _modules: Dict[str, Type[T]]
    _children: Dict[str, 'Registry']
    _parent: 'Registry'
    _base: Type[T]
    _build_func: Callable[['Registry', Config], T]

    def __getitem__(self, key: str) -> Type[T]:
        pass

    def get(self, key: str) -> Optional[Type[T]]:
        pass

    def register(
        self,
        cls: Type[T],
        *,
        name: Optional[str] = None,
        aliases: Iterable[str] = tuple(),
        force: bool = False,
    ) -> Type[T]:
        pass

    def descendent(self, name: str) -> 'Registry[T]':
        pass

    def get_descendent(self, name: str) -> Optional['Registry[T]']:
        pass

    def has_parent(self) -> bool:
        pass


class _ModulesMixin(_RegistryProto[T]):

    def __init__(self) -> None:
        self._modules: Dict[str, Type[T]] = dict()

    def __len__(self) -> int:
        return len(self._modules)

    def __contains__(self, key: str) -> bool:
        return self.get(key) is not None

    def __getitem__(self, key: str) -> Type[T]:
        item: Optional[Type[T]] = self.get(key)
        if item is None:
            raise KeyError(f'{key} does not exist in {self._name} registry')
        return item

    @property
    def modules(self) -> Dict[str, Type[T]]:
        return self._modules

    def register(
        self,
        cls: Type[T],
        *,
        name: Optional[str] = None,
        aliases: Iterable[str] = tuple(),
        force: bool = False,
    ) -> Type[T]:
        if name is None:
            name = cls.__name__
        names = [name] + list(aliases)
        if not force and len(names & self._modules.keys()) > 0:
            raise KeyError(f'{name} already exists in {self._name} registry')
        self._modules.update({name: cls for name in names})
        return cls

    def register_module(
        self,
        *,
        name: Optional[str] = None,
        aliases: Iterable[str] = tuple(),
        force: bool = False,
    ) -> Callable[[Type[T]], Type[T]]:
        return partial(
            self.register,
            name=name,
            aliases=aliases,
            force=force,
        )

    def get(self, key: str) -> Optional[Type[T]]:
        return self._modules.get(key)

    def build(self, config: Config) -> T:
        config = config.copy()
        type_ = config.pop('type')
        if isinstance(type_, str):
            type_ = self[type_]
        try:
            return type_(**config)
        except Exception as e:
            self._logger.error(
                f'Failed to build {type_.__name__} with config {config}',
            )
            raise e


class _ParentMixin(_RegistryProto[T]):

    def __init__(
        self,
        parent: Optional['Registry'] = None,
    ) -> None:
        self._children: Dict[str, 'Registry'] = dict()

        if parent is not None:
            parent._children[self._name] = cast(Registry, self)
            self._parent = parent

    @property
    def children(self) -> Dict[str, 'Registry']:
        return self._children

    @property
    def parent(self) -> 'Registry':
        return self._parent

    @property
    def root(self) -> 'Registry':
        if self.has_parent():
            return self._parent.root
        return cast(Registry, self)

    def descendent(self, name: str) -> 'Registry':
        descendent = self.get_descendent(name)
        if descendent is None:
            raise KeyError(
                f"Registry {self._name} does not have descendent named {name}."
            )
        return descendent

    def get_descendent(self, name: str) -> Optional['Registry']:
        current = self
        for child_name in name.split('.'):
            if child_name not in current._children:
                return None
            current = current._children[child_name]
        return cast(Registry, current)

    def has_parent(self) -> bool:
        return hasattr(self, '_parent')


class _BaseMixin(_RegistryProto[T]):

    def __init__(
        self,
        base: Optional[Type[T]] = None,
        register_base: bool = False,
    ) -> None:
        if base is None:
            return
        self._base = base
        if register_base:
            self.register(base)

    @property
    def base(self) -> Type[T]:
        return self._base

    def has_base(self) -> bool:
        return hasattr(self, '_base')


class _BuildFuncMixin(_RegistryProto[T]):

    def __init__(
        self,
        build_func: Optional[Callable[['Registry', Config], T]] = None,
    ) -> None:
        if build_func is not None:
            self._build_func = build_func

    @property
    def build_func(self) -> Callable[['Registry', Config], T]:
        return self._build_func

    def has_build_func(self) -> bool:
        return hasattr(self, '_build_func')


class Registry(
    _ModulesMixin[T],
    _ParentMixin[T],
    _BaseMixin[T],
    _BuildFuncMixin[T],
):

    def __init__(
        self,
        name: str,
        *,
        parent: Optional['Registry'] = None,
        base: Optional[Type[T]] = None,
        register_base: bool = False,
        build_func: Optional[Callable[['Registry', Config], T]] = None,
    ) -> None:
        assert '.' not in name
        self._name = name
        self._logger = get_logger()

        _ModulesMixin.__init__(self)
        _ParentMixin.__init__(self, parent)

        if (
            base is not None and self.has_parent() and self._parent.has_base()
            and not issubclass(base, self._parent._base)
        ):
            raise TypeError(
                f'{base} is not a subclass of {self._parent._base}',
            )
        if (base is None and self.has_parent() and self._parent.has_base()):
            base = self._parent._base
        _BaseMixin.__init__(self, base, register_base)

        if (
            build_func is None and self.has_parent()
            and self._parent.has_build_func()
        ):
            build_func = self._parent._build_func
        _BuildFuncMixin.__init__(self, build_func)

    @property
    def name(self) -> str:
        return self._name

    def register(
        self,
        cls: Type[T],
        *args,
        **kwargs,
    ) -> Type[T]:
        if self.has_base() and not issubclass(cls, self._base):
            raise TypeError(f'{cls} is not a subclass of {self._base}')
        return _ModulesMixin.register(self, cls, *args, **kwargs)

    def get(self, key: str) -> Optional[Type[T]]:
        if '.' not in key:
            return _ModulesMixin.get(self, key)
        descendent_name, key = key.rsplit('.', 1)
        descendent = self.get_descendent(descendent_name)
        if descendent is None:
            return None
        return descendent.get(key)

    def build(
        self,
        config: Union[Config, T],
        default_args: Optional[Config] = None,
    ) -> T:
        if self.has_base():
            if isinstance(config, self._base):
                return config
            if not isinstance(config, dict):
                raise TypeError(
                    f"{config.__class__.__name__} is neither dict nor "
                    f"{self._base.__name__}"
                )
        else:
            if not isinstance(config, dict):
                return config
        default_args = dict() if default_args is None else default_args.copy()
        default_args.update(config)
        if self.has_build_func():
            return self._build_func(self, default_args)
        return _ModulesMixin.build(self, default_args)


NORM_LAYERS: Registry[nn.Module] = Registry('norm layers', base=nn.Module)
NORM_LAYERS.register(nn.BatchNorm1d, name='BN1d')
NORM_LAYERS.register(nn.BatchNorm2d, name='BN2d', aliases=['BN'])
NORM_LAYERS.register(nn.BatchNorm3d, name='BN3d')
NORM_LAYERS.register(nn.SyncBatchNorm, name='SyncBN')
NORM_LAYERS.register(nn.GroupNorm, name='GN')
NORM_LAYERS.register(nn.LayerNorm, name='LN')
NORM_LAYERS.register(nn.InstanceNorm1d, name='IN1d')
NORM_LAYERS.register(nn.InstanceNorm2d, name='IN2d', aliases=['IN'])
NORM_LAYERS.register(nn.InstanceNorm3d, name='IN3d')



__all__ = [
    'MultiTeacherDistiller',
    'SingleTeacherDistiller',
]

from typing import Dict, List, Optional

import torch.nn as nn

from ..base import ModuleList, WorkflowConfig
from ..reproduction import freeze_model
from .base import DISTILLERS, BaseDistiller, DecoratorMixin


@DISTILLERS.register_module()
class MultiTeacherDistiller(DecoratorMixin, BaseDistiller):

    def __init__(
        self,
        student: nn.Module,
        online_teachers: List[nn.Module] = None,
        offline_teachers: List[nn.Module] = None,
        student_hooks: WorkflowConfig = None,
        online_teacher_hooks: Optional[Dict[int, WorkflowConfig]] = None,
        offline_teacher_hooks: Optional[Dict[int, WorkflowConfig]] = None,
        **kwargs,
    ):
        assert 'hooks' not in kwargs

        online_teachers = [] if online_teachers is None else online_teachers
        online_teacher_slice = slice(1, 1 + len(online_teachers))

        offline_teachers = [] if offline_teachers is None else offline_teachers
        for offline_teacher in offline_teachers:
            freeze_model(offline_teacher)
        offline_teacher_slice = slice(online_teacher_slice.stop, None)

        hooks = {}
        if student_hooks is not None:
            hooks[0] = student_hooks
        if online_teacher_hooks is not None:
            hooks.update({  # yapf: disable
                online_teacher_slice.start + i: hook
                for i, hook in online_teacher_hooks.items()
            })
        if offline_teacher_hooks is not None:
            hooks.update({  # yapf: disable
                offline_teacher_slice.start + i: hook
                for i, hook in offline_teacher_hooks.items()
            })

        super().__init__(
            [student] + online_teachers + offline_teachers,
            hooks=hooks,
            **kwargs,
        )
        self._online_teachers = ModuleList(online_teachers)
        self._online_teacher_slice = online_teacher_slice
        self._offline_teacher_slice = offline_teacher_slice

    @property
    def student(self) -> nn.Module:
        return self.models[0]

    @property
    def teachers(self) -> List[nn.Module]:
        return self.models[1:]

    @property
    def online_teachers(self) -> List[nn.Module]:
        return self.models[self._online_teacher_slice]

    @property
    def offline_teachers(self) -> List[nn.Module]:
        return self.models[self._offline_teacher_slice]


@DISTILLERS.register_module()
class SingleTeacherDistiller(MultiTeacherDistiller):

    def __init__(
        self,
        student: nn.Module,
        teacher: nn.Module,
        teacher_hooks: WorkflowConfig = None,
        teacher_online: bool = False,
        **kwargs,
    ):
        assert 'online_teacher_hooks' not in kwargs
        assert 'offline_teacher_hooks' not in kwargs
        assert 'online_teacher_weight_transfer' not in kwargs

        arg_prefix = ['offline', 'online'][teacher_online]
        kwargs[arg_prefix + '_teachers'] = [teacher]
        if teacher_hooks is not None:
            kwargs[arg_prefix + '_teacher_hooks'] = {0: teacher_hooks}
        super().__init__(student, **kwargs)

    @property
    def teacher(self) -> nn.Module:
        return self.teachers[0]



from typing import List, Literal

import cv2
import einops
import numpy as np
import torch
import torch.nn.functional as F

from .base import VISUALS, BaseVisual

__all__ = [
    'ActivationVisual',
]


@VISUALS.register_module()
class ActivationVisual(BaseVisual):

    def __init__(
        self,
        *args,
        mode: Literal['image', 'tensor', 'overlay'] = 'overlay',
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert mode == 'overlay', 'Only support overlay mode.'
        self._mode = mode

    def forward(
        self,
        images: List[np.ndarray],
        tensors: torch.Tensor,
    ) -> List[np.ndarray]:
        """Activation visualizer.

        Args:
            images: :math:`(N, H, W, 3)`
            tensors: :math:`(N, C, H, W)`
        """
        tensors = einops.reduce(
            tensors,
            'n c h w -> n 1 h w',
            reduction='mean',
        )
        results = []
        for i, image in enumerate(images):
            h, w, c = image.shape
            tensor: torch.Tensor = F.interpolate(
                tensors[[i]], (h, w), mode='bilinear'
            )[0, 0]
            tensor = (  # yapf: disable
                (tensor.max() - tensor)
                / (tensor.max() - tensor.min())
                * 255
            )
            tensor = tensor.detach().cpu().numpy().astype(np.uint8)
            tensor = cv2.applyColorMap(tensor, cv2.COLORMAP_JET)
            if self._mode == 'overlay':
                result = image * 0.5 + tensor * 0.5
            results.append(result)
        return results



"""Convert Logs.

Example:
    $ python -m todd.scripts.convert_logs \\
    > --reader \\
    >   type:\\'JSONReader\\' \\
    >   filepath:\\'log.json\\' \\
    > --selector \\
    >   type:\\'ContainsSelector\\' \\
    >   cond:\\'mode\\' \\
    > --selector \\
    >   type:\\'CustomSelector\\' \\
    >   cond:\\'mode==\\"train\\"\\' \\
    > --writer \\
    >   type:\\'TensorBoardWriter\\' \\
    >   filepath:\\'work_dirs/tb\\' \\
    >   main_tag:\\'mode\\' \\
    >   tag_value_dict:dict\\(loss=\\'loss\\'\\) \\
    >   global_step:\\'\\(epoch\\-1\\)\\*2\\+iter\\' \\
    >   walltime:\\'time\\'
"""

import argparse
import itertools
import numbers
import os
from abc import ABC, abstractmethod
from typing import Dict, Generator, Optional

import json
import time

from todd.base import DictAction, Registry, get_logger


class BaseReader(ABC):

    def __init__(self, filepath: str) -> None:
        super().__init__()
        self._file = open(filepath)
        self._logger = get_logger()

    def __del__(self) -> None:
        self._file.close()

    def _readline(self) -> str:
        line = self._file.readline()
        if len(line) == 0:
            raise StopIteration
        return line

    @abstractmethod
    def read(self) -> dict:
        pass

    def __iter__(self) -> Generator[dict, None, None]:
        while True:
            try:
                yield self.read()
            except StopIteration:
                break


READERS: Registry[BaseReader] = Registry('readers', base=BaseReader)


@READERS.register_module()
class TxtReader(BaseReader):

    def read(self) -> dict:
        raise NotImplementedError


@READERS.register_module()
class JSONReader(BaseReader):

    def read(self) -> dict:
        line = self._readline()
        return json.loads(line)


class BaseSelector(ABC):

    def __init__(self, cond: str) -> None:
        self._cond = cond

    @abstractmethod
    def filter(self, log: dict) -> bool:
        pass

    def __call__(self, log: dict) -> bool:
        return self.filter(log)


SELECTORS: Registry[BaseSelector] = Registry('selectors', base=BaseSelector)


@SELECTORS.register_module()
class ContainsSelector(BaseSelector):

    def filter(self, log: dict) -> bool:
        return self._cond in log


@SELECTORS.register_module()
class CustomSelector(BaseSelector):

    def filter(self, log: dict) -> bool:
        return eval(self._cond, None, log)


class BaseWriter(ABC):

    def __init__(
        self,
        filepath: str,
    ) -> None:
        self._filepath = filepath

    @abstractmethod
    def write(self, log: dict) -> None:
        pass

    def close(self) -> None:
        pass

    def __call__(self, log: dict) -> None:
        self.write(log)


WRITERS: Registry[BaseWriter] = Registry('writers', base=BaseWriter)


@WRITERS.register_module()
class TensorBoardWriter(BaseWriter):

    def __init__(
        self,
        *args,
        main_tag: Optional[str] = None,
        tag_value_dict: Dict[str, str],
        global_step: Optional[str] = None,
        walltime: Optional[str] = None,
        **kwargs,
    ) -> None:
        from torch.utils.tensorboard import SummaryWriter

        super().__init__(*args, **kwargs)
        self._main_tag = main_tag
        self._tag_value_dict = tag_value_dict
        self._global_step = global_step
        self._walltime = walltime

        self._timer = 0

        self._filepath = os.path.join(
            self._filepath,
            time.strftime('%Y%m%d_%H%M%S', time.localtime()) + '.tensorboard',
        )
        self._writer = SummaryWriter(self._filepath)

    def write(self, log: dict) -> None:
        main_tag = None if self._main_tag is None else eval(
            self._main_tag, None, log
        )
        tag_value_dict = {  # yapf: disable
            tag: eval(value, None, log)
            for tag, value in self._tag_value_dict.items()
        }
        global_step = None if self._global_step is None else eval(
            self._global_step, None, log
        )
        walltime = None if self._walltime is None else eval(
            self._walltime, None, log
        )

        self._timer += walltime or global_step or 1
        for tag, value in tag_value_dict.items():
            if main_tag is not None:
                tag = f'{main_tag}/{tag}'
            if isinstance(value, str):
                self._writer.add_text(tag, value, global_step, self._timer)
            elif isinstance(value, numbers.Number):
                self._writer.add_scalar(tag, value, global_step, self._timer)
            else:
                raise TypeError(
                    f"Unsupported data {value} with type {type(value)}"
                )

    def close(self) -> None:
        self._writer.close()


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=__doc__.replace('\n    ', '\n  '),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument('--reader', nargs='+', action=DictAction)
    parser.add_argument('--selector', nargs='+', action=DictAction)
    parser.add_argument('--writer', nargs='+', action=DictAction)
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    readers = [READERS.build(reader) for reader in args.reader]
    selectors = [SELECTORS.build(selector) for selector in args.selector]
    writers = [WRITERS.build(writer) for writer in args.writer]

    for log in itertools.chain(*readers):
        if all(selector(log) for selector in selectors):
            for writer in writers:
                writer(log)
    for writer in writers:
        writer.close()


if __name__ == '__main__':
    main()



__all__ = [
    'FocalLoss',
    'FocalWithLogitsLoss',
]

import torch

from .base import LOSSES
from .functional import BCELoss, BCEWithLogitsLoss, FunctionalLoss


class FocalMixin(FunctionalLoss):

    def __init__(
        self,
        *args,
        gamma=2.0,
        alpha=0.25,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._gamma = gamma
        self._alpha = alpha

    def forward(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        pt = (1 - pred) * target + pred * (1 - target)
        weight = self._alpha * target + (1 - self._alpha) * (1 - target)
        return super().forward(
            pred,
            target,
            weight * pt.pow(self._gamma),
            *args,
            **kwargs,
        )


@LOSSES.register_module()
class FocalLoss(FocalMixin, BCELoss):
    pass


@LOSSES.register_module()
class FocalWithLogitsLoss(FocalMixin, BCEWithLogitsLoss):
    pass



from typing import List, cast

from ..base import get_iter
from .base import SCHEDULERS, BaseScheduler

__all__ = [
    'StepScheduler',
]


@SCHEDULERS.register_module()
class StepScheduler(BaseScheduler):

    def __init__(
        self,
        *args,
        value: float = 1.0,
        iters: List[int],
        ratio: float = 0.1,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        values = [value * ratio**i for i in range(len(iters))]
        self._values = [0.0] + values
        self._iters = cast(List[float], iters) + [float('inf')]

    @property
    def value(self) -> float:
        iter_ = get_iter()
        for value, iter in zip(self._values, self._iters):
            if iter > iter_:
                return value
        raise RuntimeError(iter_, self._values, self._iters)



import argparse
import tempfile
import webbrowser
from difflib import Differ, HtmlDiff
from enum import Enum, auto
from typing import Any, Optional, Sequence

from mmcv import Config


class DiffMode(Enum):
    TEXT = auto()
    HTML = auto()

    def diff(self, a: str, b: str) -> str:
        if self == DiffMode.TEXT:
            diff = '\n'.join(Differ().compare(a, b))
        elif self == DiffMode.HTML:
            diff = HtmlDiff().make_file(a, b)
        return diff


class OutputMode(Enum):
    TERMINAL = auto()
    FILE = auto()
    BROWSER = auto()

    def output(self, diff: str, filepath: Optional[str] = None):
        if self == OutputMode.TERMINAL:
            assert filepath is None
            print(diff)
        elif self == OutputMode.FILE:
            assert filepath is not None
            with open(filepath, 'w') as f:
                f.write(diff)
        elif self == OutputMode.BROWSER:
            assert filepath is None
            with tempfile.NamedTemporaryFile(
                suffix='.html',
                delete=False,
            ) as f:
                f.write(diff.encode('utf-8'))
                webbrowser.open('file://' + f.name)


def parse_args():
    parser = argparse.ArgumentParser(description="Compare Configs")
    parser.add_argument('a')
    parser.add_argument('b')
    # TODO: add mutually exclusive parameter group
    parser.add_argument('--out')
    args = parser.parse_args()
    return args


def load_config(filepath: str) -> Sequence[str]:

    def _sort(obj: Any):
        if isinstance(obj, dict):
            return {k: _sort(obj[k]) for k in sorted(obj.keys())}
        if isinstance(obj, (list, tuple)):
            return type(obj)(_sort(e) for e in obj)
        return obj

    try:
        config = Config.fromfile(filepath)
    except Exception as e:
        print(f"Error when loading {filepath}.")
        raise e
    config = Config(
        cfg_dict=_sort(config._cfg_dict.to_dict()),
        filename=filepath,
    )
    return config.dump().split('\n')


def main():
    args = parse_args()

    a = load_config(args.a)
    b = load_config(args.b)

    if args.out is None:
        diff_mode = DiffMode.TEXT
        output_mode = OutputMode.TERMINAL
        filepath = None
    elif args.out.endswith('.txt'):
        diff_mode = DiffMode.TEXT
        output_mode = OutputMode.FILE
        filepath = args.out
    elif args.out.endswith('.html'):
        diff_mode = DiffMode.HTML
        output_mode = OutputMode.FILE
        filepath = args.out
    elif args.out == 'browser':
        diff_mode = DiffMode.HTML
        output_mode = OutputMode.BROWSER
        filepath = None
    else:
        raise ValueError(f"Unknown output mode: {args.out}.")

    diff = diff_mode.diff(a, b)
    output_mode.output(diff, filepath)


if __name__ == "__main__":
    main()



__all__ = [
    'SGR',
    'get_logger',
]

import getpass
import inspect
import logging
import os
import socket
from abc import abstractmethod
from enum import IntEnum, auto
from types import FrameType
from typing import Iterable, cast


class ANSI(IntEnum):  # fix bug in python<=3.7.1

    @classmethod
    def to_str(cls, key) -> str:
        if isinstance(key, str):
            return str(cls[key.upper()].value)
        if isinstance(key, cls):
            return str(key.value)
        return str(int(key))

    @staticmethod
    @abstractmethod
    def format(text: str) -> str:
        pass


class SGR(ANSI):
    """Select Graphic Rendition.

    Refer to https://en.wikipedia.org/wiki/ANSI_escape_code.
    """
    NORMAL = 0
    BOLD = auto()
    FAINT = auto()
    ITALIC = auto()
    UNDERLINE = auto()
    BLINK_SLOW = auto()
    BLINK_FAST = auto()
    REVERSE = auto()
    CONCEAL = auto()
    CROSSED_OUT = auto()

    FG_BLACK = 30
    FG_RED = auto()
    FG_GREEN = auto()
    FG_YELLOW = auto()
    FG_BLUE = auto()
    FG_MAGENTA = auto()
    FG_CYAN = auto()
    FG_WHITE = auto()

    BG_BLACK = 40
    BG_RED = auto()
    BG_GREEN = auto()
    BG_YELLOW = auto()
    BG_BLUE = auto()
    BG_MAGENTA = auto()
    BG_CYAN = auto()
    BG_WHITE = auto()

    @classmethod
    def format(
        cls,
        text: str,
        sgr: Iterable = tuple(),
    ) -> str:
        sgr_list = ';'.join(map(cls.to_str, sgr))
        return f"\033[{sgr_list}m{text}\033[m"


DEFAULT_FORMAT = (
    "[%(asctime)s]"
    "[Todd %(name)s]"
    "[%(filename)s:%(funcName)s:%(lineno)d]"
    " %(levelname)s: %(message)s"
)


# pragma: no cover
class Formatter(logging.Formatter):
    FORMATTERS = dict(
        zip(
            (
                logging.DEBUG,
                logging.WARNING,
                logging.ERROR,
                logging.CRITICAL,
            ),
            map(
                lambda sgr: logging.Formatter(
                    SGR.format(DEFAULT_FORMAT, sgr),
                ),
                (
                    (SGR.FAINT, ),
                    (SGR.BOLD, SGR.FG_YELLOW),
                    (SGR.BOLD, SGR.FG_RED),
                    (SGR.BOLD, SGR.BLINK_SLOW, SGR.FG_RED),
                ),
            ),
        ),
    )

    def __init__(self) -> None:
        super().__init__(DEFAULT_FORMAT)

    def format(self, record: logging.LogRecord) -> str:
        formatter = self.FORMATTERS.get(record.levelno)
        if formatter is None:
            return super().format(record)
        return formatter.format(record)


def get_logger(log_file=None) -> logging.Logger:
    frame = cast(FrameType, inspect.currentframe())
    frame = cast(FrameType, frame.f_back)
    name = frame.f_globals.get('__name__')
    logger = logging.getLogger(name)
    if not getattr(logger, '_isinitialized', False):
        setattr(logger, '_isinitialized', True)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(logging.StreamHandler())
        if log_file:
            logger.addHandler(logging.FileHandler(log_file))
        formatter = Formatter()
        for handler in logger.handlers:
            handler.setFormatter(formatter)
        logger.propagate = False
        logger.debug(
            f"logger initialized by"
            f" {getpass.getuser()}"
            f"@{socket.gethostname()}"
            f":{os.getpid()}"
        )
    return logger


if __name__ == '__main__':
    for fg_color in range(30, 38):
        for bg_color in range(40, 48):
            text = ' '.join(
                SGR.format(
                    f'{effect}:{fg_color};{bg_color}',
                    (effect, fg_color, bg_color),
                ) for effect in range(10)
            )
            print(text)



from typing import List

import torch

from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class Detach(BaseAdapt):

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.detach()


@ADAPTS.register_module()
class ListDetach(BaseAdapt):

    def forward(self, tensors: List[torch.Tensor]) -> List[torch.Tensor]:
        return [tensor.detach() for tensor in tensors]



from typing import List

import torch

from ..base import ModuleList
from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class RoIAlign(BaseAdapt):

    def __init__(self, strides: List[int], *args, **kwargs):
        import mmcv.ops

        super().__init__(*args, **kwargs)
        self._layers = ModuleList([
            mmcv.ops.RoIAlign(
                spatial_scale=1 / s,
                output_size=7,
                sampling_ratio=0,
            ) for s in strides
        ])

    def forward(
        self,
        feats: List[torch.Tensor],
        bboxes: List[torch.Tensor],
    ) -> List[torch.Tensor]:
        """
        Args:
            feats: l x n x c x h x w
            bboxes: n x r x 4

        Returns:
            roi_feats: l x r x c x 7 x 7
        """
        rois = torch.cat([  # yapf: disable
            torch.cat([b.new_full((b.shape[0], 1), i), b[:, :4]], dim=-1)
            for i, b in enumerate(bboxes)
            if b.shape[0] > 0
        ])
        roi_feats = [layer(f, rois) for layer, f in zip(self._layers, feats)]
        return roi_feats



__all__ = [
    'BaseDistiller',
    'DISTILLERS',
    'DecoratorMixin',
]

import functools
from typing import (
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Protocol,
    Type,
    TypeVar,
    cast,
    no_type_check,
)

import torch
import torch.nn as nn

from ..base import (
    Message,
    Module,
    Registry,
    Workflow,
    WorkflowConfig,
    init_iter,
)
from ..hooks import BaseHook
from ..utils import ModelLoader

T = TypeVar('T', bound='BaseDistiller')


class BaseDistiller(Module):
    DEBUG_PREFIX = '_debug_'

    @staticmethod
    def workflow_to_module(workflow: Workflow) -> nn.Module:
        module = nn.ModuleList()
        for job in workflow:
            step = job.step
            if not isinstance(step, nn.Module):
                step = nn.ModuleList(step)
            module.append(step)
        return module

    def __init__(
        self,
        models: List[nn.Module],
        *,
        hooks: Dict[int, WorkflowConfig],
        adapts: WorkflowConfig,
        losses: WorkflowConfig,
        iter_: int = 0,
        weight_transfer: Optional[Dict[str, str]] = None,
    ):
        Module.__init__(self)
        self._models = models

        self._hookflows: Dict[int, Workflow] = {  # yapf: disable
            i: Workflow.build('hooks', hooks_config)
            for i, hooks_config in hooks.items()
        }
        self._adaptflow = Workflow.build('adapts', adapts)
        self._lossflow = Workflow.build('losses', losses)

        init_iter(iter_)
        if weight_transfer is not None:
            ModelLoader.load_state_dicts(self, weight_transfer)

        for i, workflow in self._hookflows.items():
            for job in workflow:
                for step in job:
                    cast(BaseHook, step).bind(self._models[i])

        self.add_module(
            '_adapts',
            self.workflow_to_module(self._adaptflow),
        )
        self.add_module(
            '_losses',
            self.workflow_to_module(self._lossflow),
        )

    @property
    def models(self) -> List[nn.Module]:
        return self._models

    def _apply(self: T, fn: Callable[..., None]) -> T:
        for model in self._models:
            if getattr(model, 'sync_apply', True):
                model._apply(fn)
        return super()._apply(fn)

    def hookflows(self) -> Iterator[Workflow]:
        return iter(self._hookflows.values())

    def hooks(self) -> Generator[BaseHook, None, None]:
        for workflow in self.hookflows():
            for job in workflow:
                yield from job

    def track_tensors(self) -> None:
        for hook in filter(
            lambda hook: hook.tracking_mode,
            self.hooks(),
        ):
            hook.track_tensor()

    def tensors(self) -> Message:
        tensors: Message = dict()
        for job in self.hookflows():
            job(tensors)
        return tensors

    def reset(self) -> None:
        for hook in self.hooks():
            hook.reset()

    def distill(
        self,
        custom_tensors: Optional[Dict[str, torch.Tensor]] = None,
        debug: bool = False,
    ) -> Dict[str, torch.Tensor]:
        tensors = self.tensors()
        if custom_tensors is not None:
            tensors.update(custom_tensors)
        self._adaptflow(tensors)
        losses = self._lossflow(tensors.copy())

        if debug:
            tensors = {self.DEBUG_PREFIX + k: v for k, v in tensors.items()}
            return {**losses, **tensors}
        return losses


DISTILLERS: Registry[BaseDistiller] = Registry(
    'distillers',
    base=BaseDistiller,
)


class DistillerProto(Protocol):

    def __init__(self, student: nn.Module, *args, **kwargs) -> None:
        pass


DistillerType = TypeVar('DistillerType', bound=DistillerProto)


class WrapperProto(Protocol[DistillerType]):
    _distiller: DistillerType


WrapperType = TypeVar('WrapperType', bound=WrapperProto)

WrappedType = TypeVar('WrappedType')


class DecoratorMixin:

    @classmethod
    def wrap(
        cls: Type[DistillerType],
    ) -> Callable[[Type[WrappedType]], Type[WrappedType]]:

        @no_type_check
        def wrapper(wrapped_cls):

            class WrapMeta(wrapped_cls.__class__):

                def __call__(
                    wrapper_cls,
                    *args,
                    distiller,
                    **kwargs,
                ) -> WrapperType:
                    obj: WrapperType = super().__call__(*args, **kwargs)
                    obj._distiller = cls(obj, **distiller)
                    return obj

            @functools.wraps(wrapped_cls, updated=())
            class WrapClass(wrapped_cls, metaclass=WrapMeta):
                _distiller: DistillerType

                @property
                def distiller(self):
                    return self._distiller

                @property
                def sync_apply(self) -> bool:
                    return False

            return WrapClass

        return wrapper



__all__ = [
    'BaseAccessLayer',
    'ACCESS_LAYERS',
    'BaseDataset',
    'DATASETS',
]

import reprlib
from abc import abstractmethod
from enum import Enum
from typing import Any, Generic, MutableMapping, Type, TypeVar, Union

from torch.utils.data import Dataset

from ..base import Registry, get_logger

T = TypeVar('T')


class Codec(Enum):
    NONE = 'None'
    PYTORCH = 'pytorch'


class BaseAccessLayer(MutableMapping[T, Any]):

    def __init__(
        self,
        *args,
        data_root: str,
        task_name: str = '',
        codec: Union[str, Codec] = 'pytorch',
        readonly: bool = True,
        exist_ok: bool = False,
        **kwargs,
    ):
        self._data_root = data_root
        self._task_name = task_name
        self._codec = Codec(codec)
        self._readonly = readonly
        self._exist_ok = exist_ok
        self._logger = get_logger()

        self._init(*args, **kwargs)

        if readonly and not self.exists:
            raise FileNotFoundError(
                f'{self._data_root} ({self._task_name}) does not exist.',
            )
        if not readonly and self.exists and not exist_ok:
            raise FileExistsError(
                f'{self._data_root} ({self._task_name}) already exists.',
            )
        if not readonly and not self.exists:
            self.touch()

    @abstractmethod
    def _init(self, *args, **kwargs):
        pass

    @property
    @abstractmethod
    def exists(self) -> bool:
        pass

    @abstractmethod
    def touch(self):
        pass


ACCESS_LAYERS: Registry[BaseAccessLayer] = Registry(
    'access layers',
    base=BaseAccessLayer,
)


class BaseDataset(Dataset, Generic[T]):
    ACCESS_LAYER: Type = BaseAccessLayer[T]

    def __init__(self, *args, access_layer, **kwargs):
        super().__init__(*args, **kwargs)
        self._access_layer = ACCESS_LAYERS.build(
            access_layer,
            default_args=dict(type=self.ACCESS_LAYER),
        )
        self._logger = get_logger()

        self._logger.debug("Initializing keys.")
        self._keys = list(self._access_layer.keys())
        self._logger.debug(
            f"Keys {reprlib.repr(self._keys)} initialized "
            f"with length {len(self)}."
        )

    def __len__(self) -> int:
        return len(self._keys)

    def __getitem__(self, index: int) -> Any:
        key = self._keys[index]
        return self._access_layer[key]


DATASETS: Registry[BaseDataset] = Registry(
    'datasets',
    base=BaseDataset,
)



from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class Null(BaseAdapt):

    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, x):
        return x



import argparse
from typing import List

import lmdb


def parse_args():
    parser = argparse.ArgumentParser(description='Merge LMDBs')
    parser.add_argument('target_filepath')
    parser.add_argument('--source-filepaths', type=str, nargs='+')
    parser.add_argument('--dbs', type=str, nargs='+')
    args = parser.parse_args()
    return args


def merge_txns(target_txn: lmdb.Transaction, source_txn: lmdb.Transaction):
    with source_txn.cursor() as cur:
        for k, v in cur:
            target_txn.put(k, v)


def merge_dbs(
    target_env: lmdb.Environment,
    source_env: lmdb.Environment,
    dbs: List[str],
):
    for db in dbs:
        target_db = target_env.open_db(db.encode())
        source_db = source_env.open_db(db.encode())
        with target_env.begin(
            target_db, write=True,
        ) as target_txn, source_env.begin(source_db) as source_txn:
            merge_txns(target_txn, source_txn)


def merge_envs(
    target_env: lmdb.Environment,
    source_filepaths: List[str],
    dbs: List[str],
):
    for source_filepath in source_filepaths:
        with lmdb.open(
            source_filepath,
            readonly=True,
            max_dbs=len(dbs),
        ) as source_env:
            merge_dbs(target_env, source_env, dbs)


def main():
    args = parse_args()
    with lmdb.open(
        args.target_filepath,
        map_size=10 * 2**30,  # 10 GB
        readonly=False,
        max_dbs=len(args.dbs),
    ) as target_env:
        merge_envs(target_env, args.source_filepaths, args.dbs)


if __name__ == '__main__':
    main()
